# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from copy import deepcopy
import numpy as np
import pandas as pd

from pandapower.pd2ppc import _pd2ppc, _ppc2ppci
from pandapower.auxiliary import _add_auxiliary_elements, _sum_by_group
from pandapower.build_branch import get_trafo_values, _transformer_correction_factor
from pandapower.pypower.idx_bus import GS, BS, BASE_KV
from pandapower.pypower.idx_brch import BR_X, BR_R, T_BUS, F_BUS
from pandapower.pypower.idx_bus_sc import C_MAX, K_G, K_SG, V_G, \
    PS_TRAFO_IX, GS_P, BS_P, KAPPA, GS_GEN, BS_GEN
from pandapower.pypower.idx_brch_sc import K_T, K_ST

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def _get_is_ppci_bus(net, bus):
    is_bus = bus[np.isin(bus, net._is_elements_final["bus_is_idx"])]
    ppci_bus = np.unique(net._pd2ppc_lookups["bus"][is_bus])
    return ppci_bus


def _init_ppc(net):
    _check_sc_data_integrity(net)
    _add_auxiliary_elements(net)
    ppc, _ = _pd2ppc(net)

    # Init the required columns to nan
    ppc["bus"][:, [K_G, K_SG, V_G, PS_TRAFO_IX, GS_P, BS_P, GS_GEN, BS_GEN]] = np.nan
    ppc["branch"][:, [K_T, K_ST]] = np.nan

    # Add parameter K into ppc
    _add_kt(net, ppc)
    _add_gen_sc_z_kg_ks(net, ppc)
    _add_sgen_sc_z(net, ppc)
    _add_ward_sc_z(net, ppc)

    ppci = _ppc2ppci(ppc, net)

    return ppc, ppci

def _check_sc_data_integrity(net):
    if not net.gen.empty:
        for col in ("power_station_trafo", "pg_percent"):
            if col not in net.gen.columns:
                net.gen[col] = np.nan

    if not net.trafo.empty:
        for col in ("power_station_unit", "pt_percent",):
            if col not in net.trafo.columns:
                net.trafo[col] = np.nan

        for col in ("oltc", ):
            if col not in net.trafo.columns:
                net.trafo[col] = False


def _add_kt(net, ppc):
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    branch = ppc["branch"]
    # "trafo/trafo3w" are already corrected in pd2ppc, write parameter kt for trafo
    if not net.trafo.empty:
        f, t = net["_pd2ppc_lookups"]["branch"]["trafo"]
        trafo_df = net["trafo"]
        cmax = ppc["bus"][bus_lookup[get_trafo_values(trafo_df, "lv_bus")], C_MAX]
        kt = _transformer_correction_factor(trafo_df, trafo_df.vk_percent, trafo_df.vkr_percent, trafo_df.sn_mva, cmax)
        branch[f:t, K_T] = kt


def _add_ward_sc_z(net, ppc):
    for element in ("ward", "xward"):
        ward = net[element][net._is_elements_final[element]]
        if len(ward) == 0:
            continue

        ward_buses = ward.bus.values
        bus_lookup = net["_pd2ppc_lookups"]["bus"]
        ward_buses_ppc = bus_lookup[ward_buses]

        y_ward_pu = (ward["pz_mw"].values + ward["qz_mvar"].values * 1j)
        # how to calculate r and x in Ohm:
        # z_ward_pu = 1/y_ward_pu
        # vn_net = net.bus.loc[ward_buses, "vn_kv"].values
        # z_base_ohm = (vn_net ** 2)# / base_sn_mva)
        # z_ward_ohm = z_ward_pu * z_base_ohm

        buses, gs, bs = _sum_by_group(ward_buses_ppc, y_ward_pu.real, y_ward_pu.imag)
        ppc["bus"][buses, GS] += gs
        ppc["bus"][buses, BS] += bs


def _add_sgen_sc_z(net, ppc):
    # implement wind power station unit
    # implement doubly fed asynchronous generator
    # todo: what is this "_is_elements_final" thing?
    # sgen_wd = net.sgen.loc[net._is_elements_final["sgen"] & (net.sgen.generator_type=="async_doubly_fed")]
    if len(net.sgen) == 0 or "generator_type" not in net.sgen.columns:
        return

    sgen_wd = net.sgen.loc[net.sgen.in_service & (net.sgen.generator_type=="async_doubly_fed")]
    if len(sgen_wd) > 0:
        sgen_buses = sgen_wd.bus.values
        sgen_buses_ppc = net["_pd2ppc_lookups"]["bus"][sgen_buses]

        vn_net = net.bus.loc[sgen_buses, "vn_kv"].values
        base_z_ohm = vn_net ** 2 # / net.sn_mva  # by logic it must by divided by sn_mva, but why is it wrong?

        z_wd_ohm = np.sqrt(2) * sgen_wd.kappa * net.bus.loc[sgen_buses, "vn_kv"].values / (np.sqrt(3) * sgen_wd.max_ik_ka)
        z_wd_complex_ohm = (sgen_wd.rx + 1j) * z_wd_ohm / (np.sqrt(1+sgen_wd.rx**2))
        z_wd_complex_pu = z_wd_complex_ohm / base_z_ohm
        y_wd_pu = 1 / z_wd_complex_pu.values

        bus_idx, sgen_gs, sgen_bs = _sum_by_group(sgen_buses_ppc, y_wd_pu.real, y_wd_pu.imag)
        ppc['bus'][bus_idx, GS] = sgen_gs
        ppc['bus'][bus_idx, BS] = sgen_bs

        # add kappa for peak current
        ppc['bus'][sgen_buses_ppc, KAPPA] = sgen_wd.kappa.values

    sgen_g = net.sgen.loc[net.sgen.in_service & (net.sgen.generator_type=="async")]
    if len(sgen_g) > 0:
        sgen_buses = sgen_g.bus.values
        sgen_buses_ppc = net["_pd2ppc_lookups"]["bus"][sgen_buses]

        vn_net = net.bus.loc[sgen_buses, "vn_kv"].values
        base_z_ohm = vn_net ** 2 # / net.sn_mva  # by logic it must by divided by sn_mva, but why is it wrong?

        i_rg_ka = sgen_g.sn_mva / (vn_net * np.sqrt(3))
        z_g_ohm = 1 / sgen_g.lrc_pu * np.square(vn_net) / sgen_g.sn_mva
        z_g_complex_ohm = (sgen_g.rx + 1j) * z_g_ohm / (np.sqrt(1 + np.square(sgen_g.rx)))
        z_g_complex_pu = z_g_complex_ohm / base_z_ohm
        y_g_pu = 1 / z_g_complex_pu.values

        bus_idx, sgen_gs, sgen_bs = _sum_by_group(sgen_buses_ppc, y_g_pu.real, y_g_pu.imag)
        ppc['bus'][bus_idx, GS] = sgen_gs
        ppc['bus'][bus_idx, BS] = sgen_bs


def _add_gen_sc_z_kg_ks(net, ppc):
    gen = net["gen"][net._is_elements_final["gen"]]
    if len(gen) == 0:
        return
    gen_buses = gen.bus.values
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    gen_buses_ppc = bus_lookup[gen_buses]

    vn_gen = gen.vn_kv.values
    sn_gen = gen.sn_mva.values
    rdss_ohm = gen.rdss_ohm.values
    xdss_pu = gen.xdss_pu.values

    # Set to zero to avoid nan
    pg_percent = np.nan_to_num(gen.pg_percent.values)
    vn_net = net.bus.loc[gen_buses, "vn_kv"].values
    # Avoid warning by slight zero crossing caused
    sin_phi_gen = np.sqrt(np.clip(1 - gen.cos_phi.values**2, 0, None))

    # todo: division by net.sn_mva is correct but why does it cause tests to fail????
    gen_base_z_ohm = vn_net ** 2# / net.sn_mva
    r_gen, x_gen = rdss_ohm, xdss_pu * vn_gen ** 2 / sn_gen
    z_gen = (r_gen + x_gen * 1j)
    z_gen_pu = z_gen / gen_base_z_ohm
    y_gen_pu = 1 / z_gen_pu

    # this includes the Z_G from equation 21 by adding g and b to the ppc bus table
    buses, gs, bs = _sum_by_group(gen_buses_ppc, y_gen_pu.real, y_gen_pu.imag)
    ppc["bus"][buses, GS] += gs
    ppc["bus"][buses, BS] += bs

    # we need to keep track of the GS abd BS values that only come from generators
    ppc["bus"][buses, GS_GEN] = gs
    ppc["bus"][buses, BS_GEN] = bs

    # Calculate K_G
    cmax = ppc["bus"][gen_buses_ppc, C_MAX]
    # if the terminal voltage of the generator is permanently different from the nominal voltage of the generator, it may be
    # reqired to have a correction for U_{rG} as below (compare to equation 18) when calculating maximum SC current:
    kg = vn_net/(vn_gen * (1+pg_percent/100)) * cmax / (1 + xdss_pu * sin_phi_gen)
    ppc["bus"][gen_buses_ppc, K_G] = kg
    ppc["bus"][gen_buses_ppc, V_G] = vn_gen

    # Calculate G,B (r,x) on generator for peak current calculation
    r_gen_p = np.full_like(r_gen, fill_value=np.nan)
    lv_gens, hv_gens = (vn_gen <= 1.), (vn_gen > 1.)
    small_hv_gens = (sn_gen < 100) & hv_gens
    large_hv_gens = (sn_gen >= 100) & hv_gens
    if np.any(lv_gens):
        r_gen_p[lv_gens] = 0.15 * x_gen[lv_gens]
    if np.any(small_hv_gens):
        r_gen_p[small_hv_gens] = 0.07 * x_gen[small_hv_gens]
    if np.any(large_hv_gens):
        r_gen_p[large_hv_gens] = 0.05 * x_gen[large_hv_gens]
    z_gen_p = (r_gen_p + x_gen * 1j)
    z_gen_p_pu = z_gen_p / gen_base_z_ohm
    y_gen_p_pu = 1 / z_gen_p_pu

    # GS_P/GS_B will only be updated here
    _, gen_gs_p, gen_bs_p = _sum_by_group(gen_buses_ppc, y_gen_p_pu.real, y_gen_p_pu.imag)
    ppc["bus"][buses, GS_P] = gen_gs_p
    ppc["bus"][buses, BS_P] = gen_bs_p

    # Calculate K_S on power station configuration
    if np.any(~gen.power_station_trafo.isnull().values):
        f, _ = net["_pd2ppc_lookups"]["branch"]["trafo"]

        # If power station units defined with index in gen, no topological search needed
        ps_gen_mask = ~gen.power_station_trafo.isnull().values
        ps_trafo_ix = gen.loc[ps_gen_mask, "power_station_trafo"].values.astype(np.int64)
        ps_trafo = net.trafo.loc[ps_trafo_ix, :]
        _ps_trafo_real_ix =\
            pd.Series(index=net.trafo.index.values,
                      data=np.arange(net.trafo.shape[0])).loc[ps_trafo_ix].values
        ps_trafo_ppc_ix = f + _ps_trafo_real_ix
        ps_trafo_oltc_mask = ps_trafo["oltc"].values.astype(bool)
        ps_gen_buses_ppc = bus_lookup[gen.loc[ps_gen_mask, "bus"]]
        ps_cmax = ppc["bus"][ps_gen_buses_ppc, C_MAX]

        v_trafo_hv, v_trafo_lv = ps_trafo["vn_hv_kv"].values, ps_trafo["vn_lv_kv"].values
        v_q = net.bus.loc[ps_trafo["hv_bus"].values, "vn_kv"].values
        sn_trafo_mva = ps_trafo["sn_mva"].values
        Z_t = ps_trafo["vk_percent"].values / 100 * np.square(v_trafo_hv) / sn_trafo_mva
        R_t = ps_trafo["vkr_percent"].values / 100 * np.square(v_trafo_hv) / sn_trafo_mva
        X_t = np.sqrt(np.square(Z_t) - np.square(R_t))
        x_t = X_t / (np.square(v_trafo_hv) / sn_trafo_mva)
        p_t = ps_trafo["pt_percent"].values / 100

        if np.any(np.isnan(p_t)):
            # TODO: Check if tap is always on HV side
            p_t[np.isnan(p_t)] =\
                 -(ps_trafo["tap_step_percent"].values *
                   (ps_trafo["tap_max"].values - ps_trafo["tap_neutral"].values))[np.isnan(p_t)] / 100
            p_t[np.isnan(p_t)] = 0.0

        v_g = vn_gen[ps_gen_mask]
        x_g = xdss_pu[ps_gen_mask]
        p_g = pg_percent[ps_gen_mask] / 100


        if np.any(ps_trafo_oltc_mask):
            # x_g here is x''_d -> ks is correct
            # todo: if the voltage U_G is permanently higher than U_rG, then U_Gmax = U_rG*(1+p_G)
            ks = (v_q**2/v_g**2) * (v_trafo_lv**2/v_trafo_hv**2) *\
                ps_cmax / (1 + np.abs(x_g - x_t) * sin_phi_gen[ps_gen_mask])
            ppc["bus"][ps_gen_buses_ppc[ps_trafo_oltc_mask], K_SG] = ks[ps_trafo_oltc_mask]
            ppc["branch"][ps_trafo_ppc_ix[ps_trafo_oltc_mask], K_ST] = ks[ps_trafo_oltc_mask]

            # kg for sc calculation inside power station units
            kg = ps_cmax / (1 + x_g * sin_phi_gen[ps_gen_mask])
            ppc["bus"][ps_gen_buses_ppc[ps_trafo_oltc_mask], K_G] = kg[ps_trafo_oltc_mask]

        if np.any(~ps_trafo_oltc_mask):
            # (1+-p_t) in the standard
            # (1-p_t) is used if the highest partial short-circuit current of the power station unit at the high-voltage side of the unit transformer is searched for
            # if the unit transformer has no off-load taps or if no such taps are permanently used -> 1-p_t = 1
            kso = (v_q / (v_g * (1 + p_g))) * (v_trafo_lv / v_trafo_hv) * (1 - p_t) * \
                  ps_cmax / (1 + x_g * sin_phi_gen[ps_gen_mask])

            ppc["bus"][ps_gen_buses_ppc[~ps_trafo_oltc_mask], K_SG] = kso[~ps_trafo_oltc_mask]
            ppc["branch"][ps_trafo_ppc_ix[~ps_trafo_oltc_mask], K_ST] = kso[~ps_trafo_oltc_mask]

            # kg for sc calculation inside power station units
            kg = 1 / (1+p_g) * ps_cmax / (1 + x_g * sin_phi_gen[ps_gen_mask])
            ppc["bus"][ps_gen_buses_ppc[~ps_trafo_oltc_mask], K_G] = kg[~ps_trafo_oltc_mask]


def _create_k_updated_ppci(net, ppci_orig, ppci_bus, zero_sequence=False):
    ppci = deepcopy(ppci_orig)
    base_z_ohm = ppci['bus'][:, BASE_KV] ** 2 / net.sn_mva

    non_ps_gen_bus = ppci_bus[np.isnan(ppci["bus"][ppci_bus, K_SG])]
    ps_gen_bus = ppci_bus[~np.isnan(ppci["bus"][ppci_bus, K_SG])]

    ps_gen_bus_mask = ~np.isnan(ppci["bus"][:, K_SG])
    ps_trafo_mask = ~np.isnan(ppci["branch"][:, K_ST])

    if np.any(ps_trafo_mask):
        ps_trafo_ppci_ix = np.argwhere(ps_trafo_mask)
        ps_trafo_ppci_lv_bus = ppci["branch"][ps_trafo_mask, T_BUS].real.astype(np.int64)
        ps_trafo_ppci_hv_bus = ppci["branch"][ps_trafo_mask, F_BUS].real.astype(np.int64)
        ppci["bus"][np.ix_(ps_trafo_ppci_lv_bus, [PS_TRAFO_IX])] = ps_trafo_ppci_ix
        # if zero_sequence:
        #     ppci["bus"][np.ix_(ps_trafo_ppci_hv_bus, [BS])] += 1/(3 * 22 / (110 ** 2))

    if np.any(ps_gen_bus_mask):
        ppci["bus"][np.ix_(ps_gen_bus_mask, [GS_P, BS_P])] /= ppci["bus"][np.ix_(ps_gen_bus_mask, [K_SG])]
        ppci["bus"][np.ix_(ps_gen_bus_mask, [GS, BS])] += (1 / ppci["bus"][np.ix_(ps_gen_bus_mask, [K_SG])] - 1) * \
                                                          ppci["bus"][np.ix_(ps_gen_bus_mask, [GS_GEN, BS_GEN])]
        # Then, the R and X are multiplied by K_S (named K_ST here)
        ppci["branch"][np.ix_(ps_trafo_mask, [BR_X, BR_R])] *= ppci["branch"][np.ix_(ps_trafo_mask, [K_ST])]

    gen_bus_mask = np.isnan(ppci["bus"][:, K_SG]) & (~np.isnan(ppci["bus"][:, K_G]))
    if np.any(gen_bus_mask):
        ppci["bus"][np.ix_(gen_bus_mask, [GS_P, BS_P])] /= ppci["bus"][np.ix_(gen_bus_mask, [K_G])]
        ppci["bus"][np.ix_(gen_bus_mask, [GS, BS])] += (1 / ppci["bus"][np.ix_(gen_bus_mask, [K_G])] - 1) * \
                                                       ppci["bus"][np.ix_(gen_bus_mask, [GS_GEN, BS_GEN])]

    bus_ppci = {}
    if ps_gen_bus.size > 0:
        for bus in ps_gen_bus:
            ppci_gen = deepcopy(ppci)
            if not np.isfinite(ppci_gen["bus"][bus, K_SG]):
                raise UserWarning("Parameter error of K SG")
            # Correct ps gen bus
            ppci_gen["bus"][bus, [GS_P, BS_P]] /= (ppci_gen["bus"][bus, K_G] / ppci_gen["bus"][bus, K_SG])
            ppci_gen["bus"][bus, [GS, BS]] += (1 / (ppci_gen["bus"][bus, K_G]) - 1 / (ppci_gen["bus"][bus, K_SG])) * \
                                              ppci_gen["bus"][bus, [GS_GEN, BS_GEN]]

            # Correct ps transfomer
            trafo_ix = ppci_gen["bus"][bus, PS_TRAFO_IX].astype(np.int64)
            # Calculating SC inside power system unit
            ppci_gen["branch"][trafo_ix, [BR_X, BR_R]] /= ppci_gen["branch"][trafo_ix, K_ST]

            bus_ppci.update({bus: ppci_gen})

    return non_ps_gen_bus, ppci, bus_ppci

# TODO Roman: correction factor for 1ph cases




# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import copy

import numpy as np
from scipy.io import savemat

from pandapower.converter.pypower import to_ppc

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def to_mpc(net, filename=None, **kwargs):
    """
    This function converts a pandapower net to a matpower case files (.mat) version 2.
    Note: python is 0-based while Matlab is 1-based.

    INPUT:
        **net** - The pandapower net.

    OPTIONAL:
        **filename** (str, None) - File path + name of the mat file which will be created. If None
            the mpc will only be returned

        ****kwargs** - please look at to_ppc() documentation

    EXAMPLE:
        >>> from pandapower.converter import to_mpc
        >>> from pandapower.networks.power_system_test_cases import case9
        >>>
        >>> net = case9()
        >>> to_mpc(net, "case9.mat")
    """
    ppc = to_ppc(net, **kwargs)

    mpc = dict()
    mpc["mpc"] = _ppc2mpc(ppc)
    if filename is not None:
        # savemat
        savemat(filename, mpc)

    return mpc


def _ppc2mpc(ppc):
    """
    Convert network in Pypower/Matpower format
    Convert 0-based python to 1-based Matlab

    **INPUT**:
        * net - The pandapower format network
        * filename - File path + name of the mat file which is created
    """

    # convert to matpower
    # Matlab is one-based, so all entries (buses, lines, gens) have to start with 1 instead of 0
    mpc = copy.deepcopy(ppc)
    if len(np.where(mpc["bus"][:, 0] == 0)[0]):
        mpc["bus"][:, 0] = mpc["bus"][:, 0] + 1
        mpc["gen"][:, 0] = mpc["gen"][:, 0] + 1
        mpc["branch"][:, 0:2] = mpc["branch"][:, 0:2] + 1
    # adjust for the matpower converter -> taps should be 0 when there is no transformer, but are 1
    mpc["branch"][np.where(mpc["branch"][:, 8] == 1), 8] = 0
    # version is a string
    mpc["version"] = str(mpc["version"])
    # baseMVA has to be a float instead of int
    mpc["baseMVA"] = mpc["baseMVA"] * 1.0
    return mpc


if "__main__" == __name__:
    pass



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import ast
import copy
import inspect
import re
import sys
from typing import Callable, TYPE_CHECKING, Optional, Tuple

import geojson
import pandas as pd

if TYPE_CHECKING:
    from matplotlib.colors import Normalize, Colormap
from itertools import combinations
from typing_extensions import deprecated

import numpy as np
from pandas import isnull, Series, DataFrame

try:
    import matplotlib.pyplot as plt
    from matplotlib.collections import LineCollection, PatchCollection, Collection
    from matplotlib.font_manager import FontProperties
    from matplotlib.patches import Circle, Rectangle, PathPatch
    from matplotlib.textpath import TextPath
    from matplotlib.transforms import Affine2D

    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False


    class TextPath:  # so that the test does not fail
        pass

from pandapower.auxiliary import soft_dependency_error, pandapowerNet
from pandapower.plotting.patch_makers import load_patches, node_patches, gen_patches, \
    sgen_patches, ext_grid_patches, trafo_patches, storage_patches, ward_patches, xward_patches, vsc_patches
from pandapower.plotting.plotting_toolbox import _rotate_dim2, coords_from_node_geodata, \
    position_on_busbar, get_index_array

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def _get_coords_from_geojson(gj_str):
    pattern = r'"coordinates"\s*:\s*((?:\[(?:\[[^]]+],?\s*)+\])|\[[^]]+\])'
    matches = re.findall(pattern, gj_str)

    if not matches:
        return None
    if len(matches) > 1:
        raise ValueError("More than one match found in GeoJSON string")
    for m in matches:
        return ast.literal_eval(m)
    return None


class CustomTextPath(TextPath):
    """
    Create a path from the text. This class provides functionality for deepcopy, which is not
    implemented for TextPath.
    """

    def __init__(self, xy, s, size=None, prop=None, _interpolation_steps=1, usetex=False):
        """
        Create a path from the text. No support for TeX yet. Note that
        it simply is a path, not an artist. You need to use the
        PathPatch (or other artists) to draw this path onto the
        canvas.

        xy : position of the text.
        s : text
        size : font size
        prop : font property
        """
        if not MATPLOTLIB_INSTALLED:
            soft_dependency_error("class CustomTextPath", "matplotlib")
        if prop is None:
            prop = FontProperties()
        TextPath.__init__(self, xy, s, size=size, prop=prop,
                          _interpolation_steps=_interpolation_steps, usetex=usetex)
        self.s = s
        self.usetex = usetex
        self.prop = prop

    def __deepcopy__(self, memo=None):
        """
        Returns a deepcopy of the `CustomTextPath`, which is not implemented for TextPath
        """
        return self.__class__(copy.deepcopy(self._xy), copy.deepcopy(self.s), size=self.get_size(),
                              prop=copy.deepcopy(self.prop),
                              _interpolation_steps=self._interpolation_steps, usetex=self.usetex)


def create_annotation_collection(texts, coords, size, prop=None, **kwargs):
    """
    Creates PatchCollection of Texts shown at the given coordinates

    Input:
        **texts** (iterable of strings) - The texts to be

        **coords** (iterable of tuples) - The pandapower network

        **size** (int) - The pandapower network

    OPTIONAL:
        **prop** - FontProperties being passed to the TextPatches

        **kwargs** - Any other keyword-arguments will be passed to the PatchCollection.
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    tp = []
    # we convert TextPaths to PathPatches to create a PatchCollection
    if hasattr(size, "__iter__"):
        for i, t in enumerate(texts):
            tp.append(PathPatch(CustomTextPath(coords[i], t, size=size[i], prop=prop)))
    else:
        for t, c in zip(texts, coords):
            tp.append(PathPatch(CustomTextPath(c, t, size=size, prop=prop)))

    return PatchCollection(tp, **kwargs)


def add_cmap_to_collection(collection, cmap, norm, z, cbar_title, plot_colormap=True, clim=None):
    """
    Adds a colormap to the given collection.

    :param collection: collection for which to add colormap
    :type collection: matplotlib.collections.collection
    :param cmap: colormap which to use
    :type cmap: any colormap from matplotlib.colors
    :param norm: any norm which to use to translate values into colors
    :type norm: any norm from matplotlib.colors
    :param z: the array which to use in order to create the colors for the given collection
    :type z: iterable
    :param cbar_title: title of the colorbar
    :type cbar_title: str
    :param plot_colormap: flag whether the colormap is actually drawn (if False, is excluded in\
        :func:`add_single_collection`)
    :type plot_colormap: bool, default True
    :param clim: color limit of the collection
    :type clim: list(float), default None
    :return: collection - the given collection with added colormap (no copy!)
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    collection.set_cmap(cmap)
    collection.set_norm(norm)
    collection.set_array(np.ma.masked_invalid(z))
    collection.has_colormap = plot_colormap
    collection.cbar_title = cbar_title
    if clim is not None:
        collection.set_clim(clim)
    return collection


def _create_node_collection(nodes, coords, size=5, patch_type="circle", color=None, picker=False,
                            infos=None, **kwargs):
    """
    Creates a collection with patches for the given nodes. Can be used generically for different \
    types of nodes (bus in pandapower network, but also other nodes, e.g. in a networkx graph).

    :param nodes: indices of the nodes to plot
    :type nodes: iterable
    :param coords: list of node coordinates (shape (2, N))
    :type coords: iterable
    :param size: size of the patches (handed over to patch creation function)
    :type size: float
    :param patch_type: type of patches that chall be created for the nodes - can be one of\
        - "circle" for a circle\
        - "rect" for a rectangle\
        - "poly<n>" for a polygon with n edges
    :type patch_type: str, default "circle"
    :param color: colors or color of the node patches
    :type color: iterable, float
    :param picker: picker argument passed to the patch collection
    :type picker: bool, default False
    :param infos: list of infos belonging to each of the patches (can be displayed when hovering \
        over the elements)
    :type infos: list, default None
    :param kwargs: keyword arguments are passed to the patch maker and patch collection
    :type kwargs:
    :return: pc - patch collection for the nodes
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    if len(coords) == 0:
        return None

    infos = list(infos) if infos is not None else []
    patches = node_patches(coords, size, patch_type, color, **kwargs)
    pc = PatchCollection(patches, match_original=True, picker=picker)
    pc.node_indices = np.array(nodes)

    pc.patch_type = patch_type
    pc.size = size
    if 'orientation' in kwargs:
        pc.orientation = kwargs['orientation']
    if "zorder" in kwargs:
        pc.set_zorder(kwargs["zorder"])
    pc.info = infos
    pc.patch_type = patch_type
    pc.size = size
    if "zorder" in kwargs:
        pc.set_zorder(kwargs["zorder"])
    if 'orientation' in kwargs:
        pc.orientation = kwargs['orientation']

    return pc


def _create_line2d_collection(coords, indices, infos=None, picker=False, **kwargs):
    """
    Generic function to create a LineCollection from coordinates.

    :param coords: list of line coordinates (list should look like this: \
        `[[(x11, y11), (x12, y12), (x13, y13), ...], [(x21, y21), (x22, x23), ...], ...]`)
    :type coords: list or np.array
    :param indices: list of node indices
    :type indices: list or np.array
    :param infos: list of infos belonging to each of the lines (can be displayed when hovering \
        over them)
    :type infos: list, default None
    :param picker: picker argument passed to the line collection
    :type picker: bool, default False
    :param kwargs: keyword arguments are passed to the line collection
    :type kwargs:
    :return: lc - line collection for the given coordinates
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    # This would be done anyway by matplotlib - doing it explicitly makes it clear and
    # prevents unexpected behavior when observing colors being "none"
    lc = LineCollection(coords, picker=picker, **kwargs)
    lc.indices = np.array(indices)
    lc.info = infos if infos is not None else list()
    return lc


def _create_node_element_collection(node_coords, patch_maker, size=1., infos=None,
                                    repeat_infos=(1, 1), orientation=np.pi, picker=False,
                                    patch_facecolor="w", patch_edgecolor="k", line_color="k",
                                    **kwargs):
    """
    Creates matplotlib collections of node elements. All node element collections usually consist of
    one patch collection representing the element itself and a small line collection that connects
    the element to the respective node.

    :param node_coords: the coordinates (x, y) of the nodes with shape (N, 2)
    :type node_coords: iterable
    :param patch_maker: a function to generate the patches of which the collections consist (cf. \
        the patch_maker module)
    :type patch_maker: function
    :param size: patch size
    :type size: float, default 1
    :param infos: list of infos belonging to each of the elements (can be displayed when hovering \
        over them)
    :type infos: iterable, default None
    :param repeat_infos: determines how many times the info shall be repeated to match the number \
        of patches (first element) and lines (second element) returned by the patch maker
    :type repeat_infos: tuple (length 2), default (1, 1)
    :param orientation: orientation of load collection. pi is directed downwards, increasing values\
        lead to clockwise direction changes.
    :type orientation: float, default np.pi
    :param picker: picker argument passed to the line collection
    :type picker: bool, default False
    :param patch_facecolor: color of the patch face (content)
    :type patch_facecolor: matplotlib color, "w"
    :param patch_edgecolor: color of the patch edges
    :type patch_edgecolor: matplotlib color, "k"
    :param line_color: color of the connecting lines
    :type line_color: matplotlib color, "k"
    :param kwargs: key word arguments are passed to the patch function
    :type kwargs:
    :return: Return values:\
        - patch_coll - patch collection representing the element\
        - line_coll - connecting line collection

    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    angles = orientation if hasattr(orientation, '__iter__') else [orientation] * len(node_coords)
    assert len(node_coords) == len(angles), \
        "The length of coordinates does not match the length of the orientation angles!"
    if infos is None:
        infos_pc = []
        infos_lc = []
    else:
        infos_pc = list(np.repeat(infos, repeat_infos[0]))
        infos_lc = list(np.repeat(infos, repeat_infos[1]))

    linewidths = kwargs.pop("linewidths", 2.)
    linewidths = kwargs.pop("linewidth", linewidths)
    linewidths = kwargs.pop("lw", linewidths)

    lines, polys, popped_keywords = patch_maker(
        node_coords, size, angles, patch_facecolor=patch_facecolor, patch_edgecolor=patch_edgecolor,
        **kwargs)
    for kw in set(popped_keywords) & set(kwargs.keys()):
        kwargs.pop(kw)
    patch_coll = PatchCollection(polys, match_original=True, picker=picker, linewidth=linewidths,
                                 **kwargs)
    color = line_color if "color" not in kwargs else kwargs.pop("color", linewidths)
    line_coll = LineCollection(lines, color=color, picker=picker, linewidth=linewidths,
                               **kwargs)
    patch_coll.info = infos_pc
    line_coll.info = infos_lc
    return patch_coll, line_coll


def _create_complex_branch_collection(coords, patch_maker, size=1, infos=None, repeat_infos=(2, 2),
                                      picker=False, patch_facecolor="w", patch_edgecolor="k",
                                      line_color="k", linewidths=2., **kwargs):
    """
    Creates a matplotlib line collection and a matplotlib patch collection representing a branch\
    element that cannot be represented by just a line.

    :param coords: list of connecting node coordinates (usually should be \
        `[((x11, y11), (x12, y12)), ((x21, y21), (x22, y22)), ...]`)
    :type coords: (N, (2, 2)) shaped iterable
    :param patch_maker: a function to generate the patches of which the collections consist (cf. \
        the patch_maker module)
    :type patch_maker: function
    :param size: patch size
    :type size: float, default 1
    :param infos: list of infos belonging to each of the branches (can be displayed when hovering \
        over them)
    :type infos: iterable, default None
    :param repeat_infos: determines how many times the info shall be repeated to match the number \
        of patches (first element) and lines (second element) returned by the patch maker
    :type repeat_infos: tuple (length 2), default (1, 1)
    :param picker: picker argument passed to the line collection
    :type picker: bool, default False
    :param patch_facecolor: color or colors of the patch face (content)
    :type patch_facecolor: matplotlib color or iterable, "w"
    :param patch_edgecolor: color or colors of the patch edges
    :type patch_edgecolor: matplotlib color or iterable, "k"
    :param line_color: color or colors of the connecting lines
    :type line_color: matplotlib color or iterable, "k"
    :param linewidths: linewidths of the connecting lines and the patch edges
    :type linewidths: float, default 2.
    :param kwargs: key word arguments are passed to the patch maker and the patch and line \
        collections
    :type kwargs:
    :return: Return values:\
        - patch_coll - patch collection representing the branch element\
        - line_coll - line collection connecting the patches with the nodes
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    if infos is None:
        infos_pc = []
        infos_lc = []
    else:
        infos_pc = list(np.repeat(infos, repeat_infos[0]))
        infos_lc = list(np.repeat(infos, repeat_infos[1]))

    lines, patches, popped_keywords = patch_maker(
        coords, size, patch_facecolor=patch_facecolor, patch_edgecolor=patch_edgecolor,
        linewidths=linewidths, **kwargs)
    for kw in set(popped_keywords) & set(kwargs.keys()):
        kwargs.pop(kw)
    patch_coll = PatchCollection(patches, match_original=True, picker=picker, **kwargs)
    line_coll = LineCollection(lines, color=line_color, picker=picker, linewidths=linewidths,
                               **kwargs)
    patch_coll.info = infos_pc
    line_coll.info = infos_lc
    return patch_coll, line_coll


def create_bus_collection(net: pandapowerNet, buses: Optional[list]=None, size:float=5., patch_type: str="circle",
                          color=None, z=None, cmap=None, norm=None, infofunc: Optional[Callable]=None,
                          picker: bool=False, bus_geodata: Optional[pd.DataFrame]=None,
                          bus_table: str="bus", cbar_title: str="Bus Voltage [pu]", clim: Optional[Tuple[float]]=None,
                          plot_colormap: bool=True, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower buses.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **buses** (list, None) - The buses for which the collections are created.
        If None, all buses in the network are considered.

        **size** (float, 5) - patch size

        **patch_type** (str, "circle") - patch type, can be

                - "circle" for a circle
                - "rect" for a rectangle
                - "poly<n>" for a polygon with n edges

        **infofunc** (function, None) - infofunction for the patch element

        **color** (list or color, None) - color or list of colors for every element

        **z** (array, None) - array of bus voltage magnitudes for colormap. Used in case of given
        cmap. If None net.res_bus.vm_pu is used.

        **cmap** (Colormap, None) - colormap for the patch colors

        **norm** (matplotlib norm object, None) - matplotlib norm object

        **picker** (bool, False) - picker argument passed to the patch collection

        **bus_geodata** (DataFrame, None) - coordinates to use for plotting
        If None, net["bus_geodata"] is used

        **cbar_title** (str, "Bus Voltage [pu]") - colormap bar title in case of given cmap

        **clim** (tuple of floats, None) - setting the norm limits for image scaling

        **bus_table** (str, "bus") - element table to use for the buses ("bus", "bus_dc")

        **plot_colormap** (bool, True) - flag whether the colormap is actually drawn

        **kwargs** - key word arguments are passed to the patch function

    OUTPUT:
        **pc** - patch collection
    """
    if bus_table == "bus":
        dc = False
    elif bus_table == "bus_dc":
        dc = True
    else:
        raise NotImplementedError(f"bus table {bus_table} not implemented!")

    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    buses = get_index_array(buses, net[bus_table].index)
    if len(buses) == 0:
        return None
    # if bus_geodata is None:
    #     bus_geodata = net["bus_geodata"] if not dc else net["bus_dc_geodata"]

    if any(net[bus_table].loc[buses].geo.isna()):
        raise AttributeError('net.bus.geo contains NaN values, consider dropping them beforehand.')

    if bus_geodata is None:
        bus_geodata = net[bus_table].geo.apply(_get_coords_from_geojson)

    buses_with_geo = buses[np.isin(buses, bus_geodata.index.values)]
    if len(buses_with_geo) < len(buses):
        logger.warning(
            f"The following buses cannot be displayed as there is on geodata available: {set(buses) - set(buses_with_geo)}"
        )

    coords = bus_geodata.loc[buses_with_geo].values

    infos = [infofunc(bus) for bus in buses] if infofunc is not None else []

    pc = _create_node_collection(buses, coords, size, patch_type, color, picker, infos, **kwargs)

    if cmap is not None:
        if z is None:
            z = net[f"res_{bus_table}"].vm_pu.loc[buses]
        add_cmap_to_collection(pc, cmap, norm, z, cbar_title, plot_colormap, clim)

    return pc


def create_line_collection(net: pandapowerNet, lines=None,
                           line_geodata: DataFrame or None = None, bus_geodata: DataFrame or None = None,
                           use_bus_geodata: bool = False, infofunc: Callable or None = None,
                           cmap: 'Colormap' = None, norm: 'Normalize' = None, picker: bool = False,
                           z=None, cbar_title: str = "Line Loading [%]", clim=None, plot_colormap: bool = True,
                           line_table="line", **kwargs):
    """
    Creates a matplotlib line collection of pandapower lines.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **lines** (list, None) - The lines for which the collections are created. If None, all lines
        in the network are considered.

        **line_geodata** (DataFrame, None) - coordinates to use for plotting. If None,
        net.line["geo"] is used

        **bus_geodata** (DataFrame, None) - coordinates to use for plotting
        If None, net.bus["geo"] is used

        **use_bus_geodata** (bool, False) - Defines whether bus or line geodata are used.
        If False line_geodata is used, if not all lines have line geodata bus geodata is used aswell

        **infofunc** (function, None) - infofunction for the patch element

        **cmap** - colormap for the patch colors

        **norm** (matplotlib norm object, None) - matplotlib norm object

        **picker** (bool, False) - picker argument passed to the line collection

        **z** (array, None) - array of line loading magnitudes for colormap. Used in case of given
        cmap. If None net.res_line.loading_percent is used.

        **cbar_title** (str, "Line Loading [%]") - colormap bar title in case of given cmap

        **clim** (tuple of floats, None) - setting the norm limits for image scaling

        **line_table** (str, "line") - which element table to use ("line", "line_dc")

        **plot_colormap** (bool, True) - flag whether the colormap is actually drawn

        **kwargs** - key word arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection
    """
    if line_table == "line":
        dc = False
        line_geodata_table = "line_geodata"
    elif line_table == "line_dc":
        dc = True
        line_geodata_table = "line_dc_geodata"
    else:
        raise NotImplementedError(f"line table {line_table} not implemented!")

    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    if not use_bus_geodata and line_geodata is None and (
            "geo" not in net[line_table].columns or net[line_table].geo.empty):
        # if bus geodata is available, but no line geodata
        logger.warning("use_bus_geodata is automatically set to True, since net.line.geo is empty.")
        use_bus_geodata = True

    lines = get_index_array(lines, net[line_table].index)
    if len(lines) == 0:
        return None

    line_geodata: Series[str] = line_geodata.loc[lines] if line_geodata is not None else net[line_table].geo.loc[lines]
    lines_without_geo = line_geodata.index[line_geodata.isna()]

    if use_bus_geodata or not lines_without_geo.empty:
        elem_indices = lines if use_bus_geodata else lines_without_geo
        geos, line_index_successful = coords_from_node_geodata(
            element_indices=elem_indices,
            from_nodes=net.line.loc[elem_indices, 'from_bus'].values,
            to_nodes=net.line.loc[elem_indices, 'to_bus'].values,
            node_geodata=net.bus.geo,
            table_name="line",
            node_name="bus",
            ignore_no_geo_diff=True)

        line_geodata = line_geodata.combine_first(pd.Series(geos, index=line_index_successful))

    lines_without_geo = line_geodata.index[line_geodata.isna()]
    if not lines_without_geo.empty:
        logger.warning(
            f'Could not plot lines {lines_without_geo}. Bus geodata is missing for those lines!')

    infos = [infofunc(line) for line in line_geodata.index] if infofunc else []

    coords = [_get_coords_from_geojson(line_gj) for line_gj in line_geodata]

    lc = _create_line2d_collection(coords, line_geodata.index, infos, picker, **kwargs)

    if cmap is not None:
        if z is None:
            z = net[f"res_{line_table}"].loading_percent.loc[line_geodata.index]
        add_cmap_to_collection(lc, cmap, norm, z, cbar_title, plot_colormap, clim)

    return lc


def create_dcline_collection(net, dclines=None, line_geodata=None, infofunc=None, cmap=None,
                             norm=None, picker=False, z=None, cbar_title="HVDC-Line Loading [%]",
                             clim=None, plot_colormap=True, **kwargs):
    """
    Creates a matplotlib line collection of pandapower dclines.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **dclines** (list, None) - The dclines for which the collections are created. If None,
        all dclines in the network are considered.

        **bus_geodata** (DataFrame, None) - coordinates to use for plotting
        If None, net["bus_geodata"] is used

        **infofunc** (function, None) - infofunction for the patch element

        **cmap** - colormap for the patch colors

        **norm** (matplotlib norm object, None) - matplotlib norm object

        **picker** (bool, False) - picker argument passed to the line collection

        **z** (array, None) - array of line loading magnitudes for colormap. Used in case of given
        cmap. If None net.res_line.loading_percent is used.

        **cbar_title** (str, "Line Loading [%]") - colormap bar title in case of given cmap

        **clim** (tuple of floats, None) - setting the norm limits for image scaling

        **plot_colormap** (bool, True) - flag whether the colormap is actually drawn

        **kwargs** - key word arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    use_bus_geodata = True

    lines = get_index_array(dclines, net.dcline.index)
    if len(lines) == 0:
        return None

    if use_bus_geodata:
        coords, lines_with_geo = coords_from_node_geodata(
            lines, net.dcline.from_bus.loc[lines].values, net.dcline.to_bus.loc[lines].values,
            net["bus_geodata"], "line")

    if len(lines_with_geo) == 0:
        return None

    infos = [infofunc(line) for line in line_geodata.index] if infofunc else []

    lc = _create_line2d_collection(coords, lines_with_geo, infos=infos, picker=picker, **kwargs)

    loading_percent = \
        100 * net.res_dcline[["p_from_mw", "p_to_mw"]].abs().max(axis=1) / net.dcline.p_mw.abs()

    if cmap is not None:
        if z is None:
            z = loading_percent.loc[lines_with_geo]
        add_cmap_to_collection(lc, cmap, norm, z, cbar_title, plot_colormap, clim)

    return lc


def create_impedance_collection(net, impedances=None, bus_geodata=None, infofunc=None,
                                picker=False, **kwargs):
    """
    Creates a matplotlib line collection of pandapower lines.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **impedances** (list, None) - The impedances for which the collections are created.
        If None, all impedances in the network are considered.

        **bus_geodata** (DataFrame, None) - coordinates to use for plotting
        If None, net["bus_geodata"] is used

         **infofunc** (function, None) - infofunction for the patch element

        **picker** (bool, False) - picker argument passed to the line collection

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    impedances = get_index_array(impedances, net.impedance.index)
    if len(impedances) == 0:
        return None

    coords, impedances_with_geo = coords_from_node_geodata(
        impedances, net.impedance.from_bus.loc[impedances].values,
        net.impedance.to_bus.loc[impedances].values,
        bus_geodata if bus_geodata is not None else net["bus_geodata"], "impedance")

    if len(impedances_with_geo) == 0:
        return None

    infos = [infofunc(imp) for imp in impedances_with_geo] if infofunc else []

    lc = _create_line2d_collection(coords, impedances_with_geo, infos=infos, picker=picker,
                                   **kwargs)

    return lc


def create_trafo_connection_collection(net, trafos=None, bus_geodata=None, infofunc=None,
                                       cmap=None, clim=None, norm=None, z=None,
                                       cbar_title="Transformer Loading", picker=False, **kwargs):
    """
    Creates a matplotlib line collection of pandapower transformers.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **trafos** (list, None) - The transformers for which the collections are created.
        If None, all transformers in the network are considered.

        **bus_geodata** (DataFrame, None) - coordinates to use for plotting
        If None, net["bus_geodata"] is used

        **infofunc** (function, None) - infofunction for the patch element

        **cmap** - colormap for the patch colors

        **clim** (tuple of floats, None) - setting the norm limits for image scaling

        **norm** (matplotlib norm object, None) - matplotlib norm object

        **z** (array, None) - array of line loading magnitudes for colormap. Used in case of given
        cmap. If None net.res_line.loading_percent is used.

        **cbar_title** (str, "Line Loading [%]") - colormap bar title in case of given cmap

        **picker** (bool, False) - picker argument passed to the line collection

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    trafos = get_index_array(trafos, net.trafo.index)

    if bus_geodata is None:
        bus_geodata = net.bus["geo"]

    in_geodata = (net.trafo.hv_bus.loc[trafos].isin(bus_geodata.index) &
                  net.trafo.lv_bus.loc[trafos].isin(bus_geodata.index))
    trafos = trafos[in_geodata]
    trafo_table = net.trafo.loc[trafos]

    hv_geo = bus_geodata.loc[trafo_table["hv_bus"]].apply(geojson.loads).apply(geojson.utils.coords).apply(
        next).to_list()  # using next works because bus only has one coordinate pair
    lv_geo = bus_geodata.loc[trafo_table["lv_bus"]].apply(geojson.loads).apply(geojson.utils.coords).apply(
        next).to_list()
    tg = list(zip(hv_geo, lv_geo))

    info = [infofunc(tr) for tr in trafos] if infofunc is not None else []

    lc = _create_line2d_collection(tg, trafos, info, picker=picker, **kwargs)

    if cmap is not None:
        if z is None:
            z = net.res_trafo.loading_percent.loc[trafos]
        add_cmap_to_collection(lc, cmap, norm, z, cbar_title, True, clim)

    return lc


def create_trafo3w_connection_collection(net, trafos=None, bus_geodata=None, infofunc=None,
                                         **kwargs):
    """
    Creates a matplotlib line collection of pandapower 3W-transformers.
    This function can be used to create line collections for voltage fall diagrams.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **trafos** (list, None) - The 3W-transformers for which the collections are created.
        If None, all 3W-transformers in the network are considered.

        **bus_geodata** (DataFrame, None) - coordinates to use for plotting
        If None, net["bus_geodata"] is used

         **infofunc** (function, None) - infofunction for the patch element

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    trafos = get_index_array(trafos, net.trafo3w.index)

    if bus_geodata is None:
        bus_geodata = net["bus_geodata"]

    in_geodata = (net.trafo3w.hv_bus.loc[trafos].isin(bus_geodata.index) &
                  net.trafo3w.mv_bus.loc[trafos].isin(bus_geodata.index) &
                  net.trafo3w.lv_bus.loc[trafos].isin(bus_geodata.index))
    trafos = trafos[in_geodata]
    trafo_table = net.trafo3w.loc[trafos]

    hv_geo, mv_geo, lv_geo = (list(zip(*(bus_geodata.loc[trafo_table[column], var].values
                                         for var in ['x', 'y'])))
                              for column in ['hv_bus', 'mv_bus', 'lv_bus'])

    # create 3 connection lines, each of 2 points, for every trafo3w
    tg = [x for c in [list(combinations(y, 2)) for y in zip(hv_geo, mv_geo, lv_geo)] for x in c]

    # 3 times infofunc for every trafo
    info = [infofunc(x) if infofunc is not None else []
            for tr in [(t, t, t) for t in trafos]
            for x in tr]

    lc = LineCollection(tg, **kwargs)
    lc.info = info

    return lc


def create_trafo_collection(net, trafos=None, picker=False, size=None, infofunc=None, cmap=None,
                            norm=None, z=None, clim=None, cbar_title="Transformer Loading",
                            plot_colormap=True, bus_geodata=None, **kwargs):
    """
    Creates a matplotlib line collection of pandapower transformers.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **trafos** (list, None) - The transformers for which the collections are created.
        If None, all transformers in the network are considered.

        **picker** (bool, False) - picker argument passed to the patch collection

        **size** (int, None) - size of transformer symbol circles. Should be >0 and
        < 0.35*bus_distance

         **infofunc** (function, None) - infofunction for the patch element

        **kwargs** - key word arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection

        **pc** - patch collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    trafos = get_index_array(trafos, net.trafo.index)

    if bus_geodata is None:
        bus_geodata = net.bus.geo.dropna()

    in_geodata = (net.trafo.hv_bus.loc[trafos].isin(bus_geodata.index) &
                  net.trafo.lv_bus.loc[trafos].isin(bus_geodata.index))
    trafos = trafos[in_geodata]
    trafo_table = net.trafo.loc[trafos]

    coords, trafos_with_geo = coords_from_node_geodata(
        trafos, trafo_table.hv_bus.values, trafo_table.lv_bus.values, bus_geodata, "trafo")

    if len(trafos_with_geo) == 0:
        return None

    colors = kwargs.pop("color", "k")
    linewidths = kwargs.pop("linewidths", 2.)
    linewidths = kwargs.pop("linewidth", linewidths)
    linewidths = kwargs.pop("lw", linewidths)
    if cmap is not None:
        if z is None:
            z = net.res_trafo.loading_percent
        colors = [cmap(norm(z.at[idx])) for idx in trafos_with_geo]

    infos = [infofunc(i) for i in range(len(trafos_with_geo))] if infofunc is not None else []

    lc, pc = _create_complex_branch_collection(
        coords, trafo_patches, size, infos, patch_facecolor="none", patch_edgecolor=colors,
        line_color=colors, picker=picker, linewidths=linewidths, **kwargs)

    if cmap is not None:
        z_duplicated = np.repeat(z.values, 2)
        add_cmap_to_collection(lc, cmap, norm, z_duplicated, cbar_title, plot_colormap, clim)
    return lc, pc


# noinspection PyArgumentList
def create_trafo3w_collection(net, trafo3ws=None, picker=False, infofunc=None, cmap=None, norm=None,
                              z=None, clim=None, cbar_title="3W-Transformer Loading",
                              plot_colormap=True, bus_geodata=None, **kwargs):
    """
    Creates a matplotlib line collection of pandapower transformers.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **trafo3ws** (list, None) - The three winding transformers for which the collections are
        created. If None, all three winding transformers in the network are considered.

        **picker** (bool, False) - picker argument passed to the patch collection

         **infofunc** (function, None) - infofunction for the patch element

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection

        **pc** - patch collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    trafo3ws = get_index_array(trafo3ws, net.trafo3w.index)

    if bus_geodata is None:
        bus_geodata = net.bus.geo.dropna()

    in_geodata = (net.trafo3w.hv_bus.loc[trafo3ws].isin(bus_geodata.index) &
                  net.trafo3w.mv_bus.loc[trafo3ws].isin(bus_geodata.index) &
                  net.trafo3w.lv_bus.loc[trafo3ws].isin(bus_geodata.index))
    trafo3ws = trafo3ws[in_geodata]
    trafo3w_table = net.trafo3w.loc[trafo3ws]

    lines = []
    circles = []
    infos = []
    color = kwargs.pop("color", "k")
    linewidth = kwargs.pop("linewidths", 2.)
    bus_geodata = bus_geodata.apply(geojson.loads)
    if cmap is not None and z is None:
        z = net.res_trafo3w.loading_percent
    for i, idx in enumerate(trafo3w_table.index):
        # get bus geodata
        p1 = bus_geodata.loc[net.trafo3w.at[idx, "hv_bus"]].coordinates
        p2 = bus_geodata.loc[net.trafo3w.at[idx, "mv_bus"]].coordinates
        p3 = bus_geodata.loc[net.trafo3w.at[idx, "lv_bus"]].coordinates
        if np.all(p1 == p2) and np.all(p1 == p3):
            continue
        p = np.array([p1, p2, p3])
        # determine center of buses and minimum distance center-buses
        center = sum(p) / 3
        d = np.linalg.norm(p - center, axis=1)
        r = d.min() / 3
        # determine closest bus to center and vector from center to circle midpoint in closest
        # direction
        closest = d.argmin()
        to_closest = (p[closest] - center) / d[closest] * 2 * r / 3
        # determine vectors from center to circle midpoint
        order = list(range(closest, 3)) + list(range(closest))
        cm = np.empty((3, 2))
        cm[order.pop(0)] = to_closest
        ang = 2 * np.pi / 3  # 120 degree
        cm[order.pop(0)] = _rotate_dim2(to_closest, ang)
        cm[order.pop(0)] = _rotate_dim2(to_closest, -ang)
        # determine midpoints of circles
        m = center + cm
        # determine endpoints of circles
        e = (center - p) * (1 - 5 * r / 3 / d).reshape(3, 1) + p
        # save circle and line collection data
        ec = color if cmap is None else cmap(norm(z.at[idx]))
        for j in range(3):
            circles.append(Circle(m[j], r, fc=(1, 0, 0, 0), ec=ec))
            lines.append([p[j], e[j]])

        if infofunc is not None:
            infos.append(infofunc(i))
            infos.append(infofunc(i))
    if len(circles) == 0:
        return None, None
    lc = LineCollection(lines, color=color, picker=picker, linewidths=linewidth, **kwargs)
    lc.info = infos
    pc = PatchCollection(circles, match_original=True, picker=picker, linewidth=linewidth, **kwargs)
    pc.info = infos
    if cmap is not None:
        z_duplicated = np.repeat(z.values, 3)
        lc.set_cmap(cmap)
        lc.set_norm(norm)
        if clim is not None:
            lc.set_clim(clim)
        lc.set_array(np.ma.masked_invalid(z_duplicated))
        lc.has_colormap = plot_colormap
        lc.cbar_title = cbar_title
    return lc, pc


# todo geojson
def create_vsc_collection(net, vscs=None, picker=False, size=None, infofunc=None, cmap=None,
                          norm=None, z=None, clim=None, cbar_title="VSC power",
                          plot_colormap=True, bus_geodata=None, bus_dc_geodata=None, **kwargs):
    """
    Creates a matplotlib line collection of pandapower transformers.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **vscs** (list, None) - The VSC indices for which the collections are created.
        If None, all VSCs in the grid are considered.

        **picker** (bool, False) - picker argument passed to the patch collection

        **size** (int, None) - size of VSC symbol squares. Should be >0 and
        < 0.35*bus_distance

         **infofunc** (function, None) - infofunction for the patch element

        **kwargs** - keyword arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection

        **pc** - patch collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    vscs = get_index_array(vscs, net.vsc.index)

    if bus_geodata is None:
        # bus_geodata = net["bus_geodata"]
        bus_geodata = net.bus.geo.dropna()

    if bus_dc_geodata is None:
        # bus_dc_geodata = net["bus_dc_geodata"]
        bus_dc_geodata = net.bus_dc.geo.dropna()

    in_geodata = (net.vsc.bus.loc[vscs].isin(bus_geodata.index) &
                  net.vsc.bus_dc.loc[vscs].isin(bus_dc_geodata.index))
    vscs = vscs[in_geodata]
    vsc_table = net.vsc.loc[vscs]

    coords, vscs_with_geo = coords_from_node_geodata(
        vscs, vsc_table.bus.values, vsc_table.bus_dc.values, bus_geodata, "vsc", node_geodata_to=bus_dc_geodata)

    if len(vscs_with_geo) == 0:
        return None

    colors = kwargs.pop("color", "k")
    linewidths = kwargs.pop("linewidths", 2.)
    linewidths = kwargs.pop("linewidth", linewidths)
    linewidths = kwargs.pop("lw", linewidths)
    if cmap is not None:
        if z is None:
            z = net.res_vsc.p_mw
        colors = [cmap(norm(z.at[idx])) for idx in vscs_with_geo]

    infos = [infofunc(i) for i in range(len(vscs_with_geo))] if infofunc is not None else []

    lc, pc = _create_complex_branch_collection(
        coords, vsc_patches, size, infos, patch_facecolor="none", patch_edgecolor=colors,
        line_color=colors, picker=picker, linewidths=linewidths, **kwargs)

    if cmap is not None:
        z_duplicated = np.repeat(z.values, 2)
        add_cmap_to_collection(lc, cmap, norm, z_duplicated, cbar_title, plot_colormap, clim)
    return lc, pc


# todo geojson
def create_vsc_connection_collection(net, vscs=None, bus_geodata=None, bus_dc_geodata=None, infofunc=None,
                                     cmap=None, clim=None, norm=None, z=None,
                                     cbar_title="Transformer Loading", picker=False, **kwargs):
    """
    Creates a matplotlib line collection of pandapower VSCs.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **vscs** (list, None) - The VSC indices for which the collections are created.
        If None, all VSCs in the network are considered.

        **bus_geodata** (DataFrame, None) - coordinates of AC buses to use for plotting
        If None, net["bus_geodata"] is used

        **bus_dc_geodata** (DataFrame, None) - coordinates of DC buses to use for plotting
        If None, net["bus_dc_geodata"] is used

        **infofunc** (function, None) - infofunction for the patch element

        **cmap** - colormap for the patch colors

        **clim** (tuple of floats, None) - setting the norm limits for image scaling

        **norm** (matplotlib norm object, None) - matplotlib norm object

        **z** (array, None) - array of values for colormap. Used in case of given
        cmap. If None net.res_vsc.p_mw is used.

        **cbar_title** (str, "VSC active power [MW]") - colormap bar title in case of given cmap

        **picker** (bool, False) - picker argument passed to the line collection

        **kwargs - keyword arguments are passed to the patch function

    OUTPUT:
        **lc** - line collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    vscs = get_index_array(vscs, net.vsc.index)

    if bus_geodata is None:
        # bus_geodata = net["bus_geodata"]
        bus_geodata = net.bus.geo.dropna()

    if bus_dc_geodata is None:
        bus_dc_geodata = net["bus_dc_geodata"]
        bus_dc_geodata = net.bus_dc.geo.dropna()

    in_geodata = (net.vsc.bus.loc[vscs].isin(bus_geodata.index) &
                  net.vsc.bus_dc.loc[vscs].isin(bus_dc_geodata.index))
    vscs = vscs[in_geodata]
    vsc_table = net.vsc.loc[vscs]

    ac_geo = list(zip(bus_geodata.loc[vsc_table["bus"], "x"].values,
                      bus_geodata.loc[vsc_table["bus_dc"], "y"].values))
    dc_geo = list(zip(bus_dc_geodata.loc[vsc_table["bus_dc"], "x"].values,
                      bus_dc_geodata.loc[vsc_table["bus_dc"], "y"].values))
    vg = list(zip(ac_geo, dc_geo))

    info = [infofunc(v) for v in vscs] if infofunc is not None else []

    lc = _create_line2d_collection(vg, vscs, info, picker=picker, **kwargs)

    if cmap is not None:
        if z is None:
            z = net.res_vsc.p_mw.loc[vscs]
        add_cmap_to_collection(lc, cmap, norm, z, cbar_title, True, clim)

    return lc


@deprecated("Busbar geodata is no longer supported for plotting geodata.")
def create_busbar_collection(net, buses=None, infofunc=None, cmap=None, norm=None, picker=False,
                             z=None, cbar_title="Bus Voltage [p.u.]", clim=None, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower buses plotted as busbars

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **buses** (list, None) - The buses for which the collections are created. If None, all buses
        which have the entry coords in bus_geodata are considered.

        **line_geodata** (DataFrame, None) - coordinates to use for plotting. If None,
        net["line_geodata"] is used

        **infofunc** (function, None) - infofunction for the line element

        **cmap** - colormap for the line colors

        **norm** (matplotlib norm object, None) - matplotlib norm object

        **picker** (bool, False) - picker argument passed to the patch collection

        **z** (array, None) - array of line loading magnitudes for colormap. Used in case of given
        cmap. If None net.res_line.loading_percent is used.

        **cbar_title** (str, "Line Loading [%]") - colormap bar title in case of given cmap

        **clim** (tuple of floats, None) - setting the norm limits for image scaling

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **bbc** - busbar collection
    """

    if buses is None:
        buses = net.bus_geodata.loc[~isnull(net.bus_geodata.coords)].index

    if cmap is not None:
        # determine color of busbar by vm_pu
        if z is None and net is not None:
            z = net.res_bus.vm_pu.loc[buses]
        else:
            logger.warning("z is None and no net is provided")

    # the busbar is just a line collection with coords from net.bus_geodata
    lc = create_line_collection(net, lines=buses, line_geodata=net.bus["geo"], bus_geodata=None,
                                norm=norm, cmap=cmap, infofunc=infofunc, picker=picker, z=z,
                                cbar_title=cbar_title, clim=clim, **kwargs)

    return lc


def create_load_collection(net, loads=None, size=1., infofunc=None, orientation=np.pi, picker=False,
                           **kwargs):
    """
    Creates a matplotlib patch collection of pandapower loads.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **loads** (list of ints, None) - the loads to include in the collection

        **size** (float, 1) - patch size

        **infofunc** (function, None) - infofunction for the patch element

        **orientation** (float, np.pi) - orientation of load collection. pi is directed downwards,
        increasing values lead to clockwise direction changes.

        **picker** (bool, False) - picker argument passed to the patch collectionent

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **load_pc** - patch collection

        **load_lc** - line collection
    """
    loads = get_index_array(loads, net.load.index)
    infos = [infofunc(i) for i in range(len(loads))] if infofunc is not None else []
    node_coords = net.bus.loc[net.load.loc[loads, "bus"].values, "geo"].apply(geojson.loads).apply(
        geojson.utils.coords).apply(next).to_list()

    color = kwargs.pop("color", "k")

    load_pc, load_lc = _create_node_element_collection(
        node_coords, load_patches, size=size, infos=infos, orientation=orientation,
        picker=picker, line_color=color, **kwargs)
    return load_pc, load_lc


def create_gen_collection(net, gens=None, size=1., infofunc=None, orientation=np.pi, picker=False,
                          **kwargs):
    """
    Creates a matplotlib patch collection of pandapower gens.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **gens** (list of ints, None) - the generators to include in the collection

        **size** (float, 1) - patch size

        **infofunc** (function, None) - infofunction for the patch element

        **orientation** (float or list of floats, np.pi) - orientation of gen collection. pi is\
        directed downwards, increasing values lead to clockwise direction changes.

        **picker** (bool, False) - picker argument passed to the patch collectionent

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **gen_pc** - patch collection

        **gen_lc** - line collection
    """
    gens = get_index_array(gens, net.gen.index)
    infos = [infofunc(i) for i in range(len(gens))] if infofunc is not None else []
    node_coords = net.bus.loc[net.gen.loc[gens, "bus"].values, "geo"].apply(geojson.loads).apply(
        geojson.utils.coords).apply(next).to_list()

    color = kwargs.pop("color", "k")

    gen_pc, gen_lc = _create_node_element_collection(
        node_coords, gen_patches, size=size, infos=infos, orientation=orientation,
        picker=picker, line_color=color, **kwargs)
    return gen_pc, gen_lc


def create_sgen_collection(net, sgens=None, size=1., infofunc=None, orientation=np.pi, picker=False,
                           **kwargs):
    """
    Creates a matplotlib patch collection of pandapower sgen.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **sgens** (list of ints, None) - the static generators to include in the collection

        **size** (float, 1) - patch size

        **infofunc** (function, None) - infofunction for the patch elem

        **picker** (bool, False) - picker argument passed to the patch collectionent

        **orientation** (float, np.pi) - orientation of static generator collection. pi is directed\
        downwards, increasing values lead to clockwise direction changes.

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **sgen_pc** - patch collection

        **sgen_lc** - line collection
    """
    sgens = get_index_array(sgens, net.sgen.index)
    infos = [infofunc(i) for i in range(len(sgens))] if infofunc is not None else []
    node_coords = net.bus.loc[net.sgen.loc[sgens, "bus"].values, "geo"].apply(geojson.loads).apply(
        geojson.utils.coords).apply(next).to_list()

    color = kwargs.pop("color", "k")

    sgen_pc, sgen_lc = _create_node_element_collection(
        node_coords, sgen_patches, size=size, infos=infos, orientation=orientation,
        picker=picker, line_color=color, **kwargs)
    return sgen_pc, sgen_lc


def create_storage_collection(net, storages=None, size=1., infofunc=None, orientation=np.pi,
                              picker=False, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower storage element.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **storages** (list of ints, None) - the net.storage.index values to include in the
        collection

        **size** (float, 1) - patch size

        **infofunc** (function, None) - info function for the patch element

        **picker** (bool, False) - picker argument passed to the patch collection

        **orientation** (float, np.pi) - orientation of static generator collection. pi is directed\
        downwards, increasing values lead to clockwise direction changes.

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **storage_pc** - patch collection

        **storage_lc** - line collection
    """
    infos = [infofunc(i) for i in range(len(storages))] if infofunc is not None else []
    node_coords = net.bus.loc[net.storage.loc[storages, "bus"].values, "geo"].apply(geojson.loads).apply(
        geojson.utils.coords).apply(next).to_list()

    color = kwargs.pop("color", "k")

    storage_pc, storage_lc = _create_node_element_collection(
        node_coords, storage_patches, size=size, infos=infos, orientation=orientation,
        picker=picker, line_color=color, **kwargs)
    return storage_pc, storage_lc


def create_ext_grid_collection(net, ext_grids=None, size=1., infofunc=None, orientation=0,
                               picker=False, ext_grid_buses=None, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower ext_grid. Parameters
    ext_grids, ext_grid_buses can be used to specify, which ext_grids the collection should be
    created for.

    Input:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **size** (float, 1) - patch size

        **infofunc** (function, None) - infofunction for the patch element

        **orientation** (float, 0) - orientation of load collection. 0 is directed upwards,
        increasing values lead to clockwise direction changes.

        **picker** (bool, False) - picker argument passed to the patch collection

        **ext_grid_buses** (np.ndarray, None) - buses to be used as ext_grid locations

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **ext_grid1** - patch collection

        **ext_grid2** - patch collection
    """
    ext_grids = get_index_array(ext_grids, net.ext_grid.index)
    if ext_grid_buses is None:
        ext_grid_buses = net.ext_grid.bus.loc[ext_grids].values
    else:
        assert len(ext_grids) == len(ext_grid_buses), \
            "Length mismatch between chosen ext_grids and ext_grid_buses."
    infos = [infofunc(ext_grid_idx) for ext_grid_idx in ext_grids] if infofunc is not None else []

    # This code does not support bus bars. It would require the format here to be a bit different.
    node_coords = net.bus.geo.loc[ext_grid_buses].apply(geojson.loads).apply(geojson.utils.coords).apply(
        next).values.tolist()

    color = kwargs.pop("color", "k")

    ext_grid_pc, ext_grid_lc = _create_node_element_collection(
        node_coords, ext_grid_patches, size=size, infos=infos, orientation=orientation,
        picker=picker, hatch='XXX', line_color=color, **kwargs)

    return ext_grid_pc, ext_grid_lc


def create_line_switch_collection(net, switches=None, size=1, distance_to_bus=3, use_line_geodata=False, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower line-bus switches.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:

        **size** (float, 1) - Size of the switch patches

        **distance_to_bus** (float, 3) - Distance of the switch patch from the bus patch

        **use_line_geodata** (bool, False) - If True, line coordinates are used to identify the
        switch position

        **switch_index** (list, []) - Possibility to create line switch collections with a subset of switches in net.switch.index.
        If left empty, all switches are taken into the line switch collection.


        **kwargs - Key word arguments are passed to the patch function

    OUTPUT:
        **switches** - patch collection
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")

    if switches is None:
        switches = net.switch.index[net.switch.et == "l"]  # only line switches

    color = kwargs.pop("color", "k")

    switch_patches = []
    for switch in switches:
        sb = net.switch.bus.loc[switch]
        line = net.line.loc[net.switch.element.loc[switch]]
        fb = line.from_bus
        tb = line.to_bus

        line_buses = {fb, tb}
        target_bus = list(line_buses - {sb})[0]

        if sb not in net.bus.index or target_bus not in net.bus.index:
            logger.warning(f"Bus coordinates for switch {switch} not found, skipped switch!")
            continue

        # switch bus and target coordinates
        pos_sb = np.array(next(geojson.utils.coords(geojson.loads(net.bus.loc[sb, "geo"]))))
        pos_tb = np.zeros(2)

        use_bus_geodata = False

        if use_line_geodata:
            if line.name in net.line.geo.dropna().index:
                line_coords = net.line.loc[line.name, "geo"]
                # check, which end of the line is nearer to the switch bus
                intersection = position_on_busbar(net, sb, busbar_coords=line_coords)
                if intersection is not None:
                    pos_sb = intersection
                if len(line_coords) >= 2:
                    if abs(line_coords[0][0] - pos_sb[0]) < 0.01 and \
                            abs(line_coords[0][1] - pos_sb[1]) < 0.01:
                        pos_tb = np.array([line_coords[1][0], line_coords[1][1]])
                    else:
                        pos_tb = np.array([line_coords[-2][0], line_coords[-2][1]])
                else:
                    use_bus_geodata = True
            else:
                use_bus_geodata = True

        if not use_line_geodata or use_bus_geodata:
            pos_tb = np.array(next(geojson.utils.coords(geojson.loads(net.bus.loc[target_bus, "geo"]))))

        # position of switch symbol
        vec = pos_tb - pos_sb
        mag = np.linalg.norm(vec)
        pos_sw = pos_sb + vec / mag * distance_to_bus

        # rotation of switch symbol
        angle = np.arctan2(vec[1], vec[0])
        rotation = Affine2D().rotate_around(pos_sw[0], pos_sw[1], angle)

        # color switch by state
        col = color if net.switch.closed.loc[switch] else "white"

        # create switch patch (switch size is respected to center the switch on the line)
        patch = Rectangle((pos_sw[0] - size / 2, pos_sw[1] - size / 2), size, size, facecolor=col, edgecolor=color)
        # apply rotation
        patch.set_transform(rotation)

        switch_patches.append(patch)

    switches = PatchCollection(switch_patches, match_original=True, **kwargs)
    return switches


def create_bus_bus_switch_collection(net, size=1., helper_line_style=':', helper_line_size=1.,
                                     helper_line_color="gray", switches=None, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower bus-bus switches. Switches are plotted in
    the center between two buses with a "helper" line (dashed and thin) being drawn between the
    buses as well.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:

        **size** (float, 1.0) - Size of the switch patches

        **helper_line_style** (string, ':') - Line style of the "helper" line being plotted between
        two buses connected by a bus-bus switch

        **helper_line_size** (float, 1.0) - Line width of the "helper" line being plotted between
        two buses connected by a bus-bus switch

        **helper_line_color** (string, "gray") - Line color of the "helper" line being plotted
        between two buses connected by a bus-bus switch

        **switches** (list, []) - switches to include in the collection

        **kwargs - Key word arguments are passed to the patch function

    OUTPUT:
        **switches**, **helper_lines** - tuple of patch collections
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    if switches is None:
        switches = net.switch.index.to_list()
    lbs_switches = net.switch.index[(net.switch.et == "b") & (net.switch.index.isin(switches))]
    color = kwargs.pop("color", "k")
    switch_patches = []
    line_patches = []
    for switch in lbs_switches:
        switch_bus = net.switch.bus.loc[switch]
        target_bus = net.switch.element.loc[switch]
        if switch_bus not in net.bus.index or target_bus not in net.bus.index \
                or pd.isnull(net.bus.loc[switch_bus, "geo"]) or pd.isnull(net.bus.loc[target_bus, "geo"]):
            logger.warning("Bus coordinates for switch %s not found, skipped switch!" % switch)
            continue
        # switch bus and target coordinates
        pos_sb = np.array(next(geojson.utils.coords(geojson.loads(net.bus.loc[switch_bus, "geo"]))))
        pos_tb = np.array(next(geojson.utils.coords(geojson.loads(net.bus.loc[target_bus, "geo"]))))
        # position of switch symbol
        vec = pos_tb - pos_sb
        pos_sw = pos_sb + vec * 0.5 if not np.allclose(pos_sb, pos_tb) else pos_tb
        # rotation of switch symbol
        angle = np.arctan2(vec[1], vec[0])
        rotation = Affine2D().rotate_around(pos_sw[0], pos_sw[1], angle)
        # color switch by state
        col = color if net.switch.closed.loc[switch] else "white"
        # create switch patch (switch size is respected to center the switch on the line)
        patch = Rectangle((pos_sw[0] - size / 2, pos_sw[1] - size / 2), size, size, facecolor=col,
                          edgecolor=color)
        # apply rotation
        patch.set_transform(rotation)
        # add to collection lists
        switch_patches.append(patch)
        line_patches.append([pos_sb.tolist(), pos_tb.tolist()])
    # create collections and return
    switches = PatchCollection(switch_patches, match_original=True, **kwargs)
    helper_lines = LineCollection(line_patches, linestyles=helper_line_style,
                                  linewidths=helper_line_size, colors=helper_line_color)
    return switches, helper_lines


def create_ward_collection(net, wards=None, ward_buses=None, size=5., bus_geodata=None, infofunc=None, picker=False,
                           orientation=0, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower wards. Wards are plotted as a grounded impedance.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:

        **wards** (list of ints, None) - the wards to include in the collection

        **ward_buses** (list of ints, None) - the buses connected to the wards

        **size** (float, 1) - patch size

        **infofunc** (function, None) - infofunction for the patch elem

        **picker** (bool, False) - picker argument passed to the patch collectionent

        **orientation** (float, np.pi) - orientation of static generator collection. pi is directed\
        downwards, increasing values lead to clockwise direction changes.

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **ward_pc** - patch collection

        **ward_lc** - line collection
    """
    wards = get_index_array(wards, net.ward.index)
    if ward_buses is None:
        ward_buses = net.ward.bus.loc[wards].values
    else:
        assert len(wards) == len(ward_buses), \
            "Length mismatch between chosen xwards and xward_buses."
    infos = [infofunc(i) for i in range(len(wards))] if infofunc is not None else []
    node_coords = net.bus.geo.loc[ward_buses].apply(geojson.loads).apply(geojson.utils.coords).apply(
        next).values.tolist()

    color = kwargs.pop("color", "k")

    ward_pc, ward_lc = _create_node_element_collection(
        node_coords, ward_patches, size=size, infos=infos, orientation=orientation,
        picker=picker, line_color=color, **kwargs)  # patch_facecolor=color, patch_edgecolor=color
    return ward_pc, ward_lc


def create_xward_collection(net, xwards=None, xward_buses=None, size=5., bus_geodata=None, infofunc=None, picker=False,
                            orientation=0, **kwargs):
    """
    Creates a matplotlib patch collection of pandapower xwards. Extended wards are plotted as a grounded impedance with
    a generator.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:

        **xwards** (list of ints, None) - the wards to include in the collection

        **xward_buses** (list of ints, None) - the buses connected to the wards

        **size** (float, 1) - patch size

        **infofunc** (function, None) - infofunction for the patch elem

        **picker** (bool, False) - picker argument passed to the patch collectionent

        **orientation** (float, np.pi) - orientation of static generator collection. pi is directed\
        downwards, increasing values lead to clockwise direction changes.

        **kwargs - key word arguments are passed to the patch function

    OUTPUT:
        **ward_pc** - patch collection

        **ward_lc** - line collection
    """
    xwards = get_index_array(xwards, net.xward.index)
    if xward_buses is None:
        xward_buses = net.xward.bus.loc[xwards].values
    else:
        assert len(xwards) == len(xward_buses), \
            "Length mismatch between chosen xwards and xward_buses."
    infos = [infofunc(i) for i in range(len(xwards))] if infofunc is not None else []
    node_coords = net.bus.geo.loc[xward_buses].apply(geojson.loads).apply(geojson.utils.coords).apply(
        next).values.tolist()

    color = kwargs.pop("color", "w")

    xward_pc, xward_lc = _create_node_element_collection(
        node_coords, xward_patches, size=size, infos=infos, orientation=orientation,
        picker=picker, line_color=color, **kwargs)
    return xward_pc, xward_lc


def draw_collections(collections, figsize=(10, 8), ax=None, plot_colorbars=True, set_aspect=True,
                     axes_visible=(False, False), copy_collections=True, draw=True, aspect=('equal', 'datalim'),
                     autoscale=(True, True, True)):
    """
    Draws matplotlib collections which can be created with the create collection functions.

    Input:
        **collections** (list) - iterable of collection objects, may include tuples of collections

    OPTIONAL:
        **figsize** (tuple, (10,8)) - figsize of the matplotlib figure

        **ax** (axis, None) - matplotlib axis object to plot into, new axis is created if None

        **plot_colorbars** (bool, True) - defines whether colorbars should be plotted

        **set_aspect** (bool, True) - defines whether 'equal' and 'datalim' aspects of axis scaling
        should be set.

        **axes_visible** (tuple, (False, False)) - defines visibility of (xaxis, yaxis)

    OUTPUT:
        **ax** - matplotlib axes
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    if ax is None:
        plt.figure(facecolor="white", figsize=figsize)
        plt.subplots_adjust(left=0.01, right=0.99, top=0.99, bottom=0.05,
                            wspace=0.02, hspace=0.04)
    ax = ax or plt.gca()

    add_collections_to_axes(ax, collections, plot_colorbars=plot_colorbars,
                            copy_collections=copy_collections)

    try:
        ax.set_facecolor("white")
    except:
        ax.set_axis_bgcolor("white")
    ax.xaxis.set_visible(axes_visible[0])
    ax.yaxis.set_visible(axes_visible[1])
    if not any(axes_visible):
        # removes bounding box of the plot also
        ax.axis("off")
    if set_aspect:
        ax.set_aspect(aspect[0], aspect[1])
    ax.autoscale_view(autoscale[0], autoscale[1], autoscale[2])
    ax.margins(.02)
    if draw:
        plt.draw()
    return ax


def add_single_collection(c, ax, plot_colorbars, copy_collections):
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error("add_single_collection()", "matplotlib")
    if copy_collections:
        c = copy.deepcopy(c)
    ax.add_collection(c)
    if plot_colorbars and hasattr(c, "has_colormap") and c.has_colormap:
        extend = c.extend if hasattr(c, "extend") else "neither"
        cbar_load = plt.colorbar(c, extend=extend, ax=ax)
        if hasattr(c, "cbar_title"):
            cbar_load.ax.set_ylabel(c.cbar_title)


def add_collections_to_axes(ax, collections, plot_colorbars=True, copy_collections=True):
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error("add_collections_to_axes()", "matplotlib")
    for i, c in enumerate(collections):
        if Collection in inspect.getmro(c.__class__):
            # if Collection is in one of the base classes of c
            add_single_collection(c, ax, plot_colorbars, copy_collections)
        elif isinstance(c, tuple) or isinstance(c, list):
            # if c is a tuple or a list of collections
            add_collections_to_axes(ax, c, plot_colorbars, copy_collections)
        else:
            logger.warning("{} in collections is of unknown type. Skipping".format(i))


if __name__ == "__main__":
    # if 0:
    #     from pandapower.create import create_empty_network, create_bus, create_gen, create_load, create_ext_grid, /
    #         create_transformer
    #
    #     ntw = pp.create_empty_network()
    #     b1 = create_bus(ntw, 10, geodata=(5, 10))
    #     b2 = create_bus(ntw, 0.4, geodata=(5, 15))
    #     b3 = create_bus(ntw, 0.4, geodata=(0, 22))
    #     b4 = create_bus(ntw, 0.4, geodata=(8, 20))
    #     create_gen(ntw, b1, p_mw=0.1)
    #     create_load(ntw, b3, p_mw=0.1)
    #     create_ext_grid(ntw, b4)
    #
    #     create_line(ntw, b2, b3, 2.0, std_type="NAYY 4x50 SE")
    #     create_line(ntw, b2, b4, 2.0, std_type="NAYY 4x50 SE")
    #     create_transformer(ntw, b1, b2, std_type="0.63 MVA 10/0.4 kV")
    #     create_transformer(ntw, b3, b4, std_type="0.63 MVA 10/0.4 kV")
    #
    #     bus_col = create_bus_collection(ntw, size=0.2, color="k")
    #     line_col = create_line_collection(ntw, use_line_geodata=False, color="k", linewidth=3.)
    #     lt, bt = create_trafo_collection(ntw, size=2, linewidth=3.)
    #     load_col1, load_col2 = create_load_collection(ntw, linewidth=2.,
    #                                                   infofunc=lambda x: ("load", x))
    #     gen1, gen2 = create_gen_collection(ntw, linewidth=2.,
    #                                        infofunc=lambda x: ("gen", x))
    #     eg1, eg2 = create_ext_grid_collection(ntw, size=2.,
    #                                           infofunc=lambda x: ("ext_grid", x))
    #
    #     draw_collections([bus_col, line_col, load_col1, load_col2, gen1, gen2, lt, bt, eg1, eg2])
    # else:
    #     pass
    pass



from pandapower.toolbox.power_factor import *
from pandapower.toolbox.result_info import *
from pandapower.toolbox.element_selection import *
from pandapower.toolbox.data_modification import *
from pandapower.toolbox.grid_modification import *
from pandapower.toolbox.comparison import *



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import sys
import time

import numpy as np
import pandas as pd
from pandas import Index

from pandapower.auxiliary import soft_dependency_error, ensure_iterability
from pandapower.control.util.characteristic import SplineCharacteristic, Characteristic

try:
    import matplotlib.pyplot as plt

    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False

try:
    import pandaplan.core.pplog as pplog
except ImportError:
    import logging as pplog

logger = pplog.getLogger(__name__)


def asarray(val, dtype=np.float64):
    """
    make sure the value is a numpy array of a certain dtype
    in contrast to np.asarray, it also converts a scalar value to an array,
    e.g. asarray(0) = array([0])
    :param val: scalar or list-like or pandas index input
    :param dtype: np.dtype for the output array
    :return: numpy array of the dtype set
    """
    # val is a list, tuple etc
    if not np.isscalar(val) and np.ndim(val) > 0:
        np_val = np.asarray(val, dtype=dtype)
    else:
        # val is a scalar number
        np_val = np.asarray([val], dtype=dtype)

    return np_val


def get_controller_index_by_type(net, ctrl_type, idx=[]):
    """
    Returns controller indices of a given type as list.
    """
    idx = idx if len(idx) else net.controller.index
    is_of_type = net.controller.object.apply(lambda x: isinstance(x, ctrl_type))
    return list(net.controller.index.values[net.controller.index.isin(idx) & is_of_type])


def get_controller_index_by_typename(net, typename, idx=[], case_sensitive=False):
    """
    Returns controller indices of a given name of type as list.
    """
    idx = idx if len(idx) else net.controller.index
    if case_sensitive:
        return [i for i in idx if str(net.controller.object.at[i]).split(" ")[0] == typename]
    else:
        return [i for i in idx if str(net.controller.object.at[i]).split(" ")[0].lower() ==
                typename.lower()]


def _controller_attributes_query(controller, parameters):
    """
    Returns a boolean if the controller attributes matches given parameter dict data
    """
    complete_match = True
    element_index_match = True
    for key in parameters.keys():
        if key not in controller.__dict__:
            logger.debug(str(key) + " is no attribute of controller object " + str(controller))
            return False
        try:
            match = bool(controller.__getattribute__(key) == parameters[key])
        except ValueError:
            try:
                match = all(controller.__getattribute__(key) == parameters[key])
            except ValueError:
                match = bool(len(set(controller.__getattribute__(key)) & set(parameters[key])))
        if key == "element_index":
            element_index_match = match
        else:
            complete_match &= match

    if complete_match and not element_index_match:
        intersect_elms = set(ensure_iterability(controller.__getattribute__("element_index"))) & \
                         set(ensure_iterability(parameters["element_index"]))
        if len(intersect_elms):
            logger.debug("'element_index' has an intersection of " + str(intersect_elms) +
                         " with Controller %i" % controller.index)

    return complete_match & element_index_match


def get_controller_index(net, ctrl_type=None, parameters=None, idx=[]):
    """ Returns indices of searched controllers. Parameters can specify the search query.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **controller_type** (controller object or string name of controller object)

        **parameters** (None, dict) - Dict of parameter names, which are in the controller object or
            net.controller DataFrame

        **idx** ([], list) - list of indices in net.controller to be searched for. If list is empty
            all indices are considered.

    OUTPUT:
        **idx** (list) - index / indices of controllers in net.controller which are in idx and
            matches given ctrl_type or parameters
    """
    #    logger.debug(ctrl_type, parameters, idx)
    idx = idx if len(idx) else net.controller.index
    if ctrl_type is not None:
        if isinstance(ctrl_type, str):
            idx = get_controller_index_by_typename(net, ctrl_type, idx)
        else:
            idx = get_controller_index_by_type(net, ctrl_type, idx)
    if isinstance(parameters, dict):
        df_keys = [k for k in parameters.keys() if k in net.controller.columns]
        attributes_keys = list(set(parameters.keys()) - set(df_keys))
        attributes_dict = {k: parameters[k] for k in attributes_keys}
        # query of parameters in net.controller dataframe
        idx = Index(idx, dtype=np.int64)
        for df_key in df_keys:
            idx = idx.intersection(net.controller.index[net.controller[df_key] == parameters[df_key]])
        # query of parameters in controller object attributes
        matches = net.controller.object.apply(lambda ctrl: _controller_attributes_query(ctrl, attributes_dict))
        idx = list(net.controller.index.values[net.controller.index.isin(idx) & matches])
    return idx


def log_same_type_existing_controllers(net, this_ctrl_type, index=None, matching_params=None,
                                       **kwargs):
    """
    Logs same type controllers, if a controller is created.
    INPUT:
        **net** - pandapower net

        **this_ctrl_type** (controller object or string name of controller object)

    OPTIONAL:
        **index** (int) - index in net.controller of the controller to be created

        **matching_params** (dict) - parameters, which must be equal if same type controller should
            be logged.

        ****kwargs** - unused arguments, given to avoid unexpected input arguments
    """
    index = str(index)
    if isinstance(matching_params, dict):
        same_type_existing_ctrl = get_controller_index(net, ctrl_type=this_ctrl_type,
                                                       parameters=matching_params)
        if len(same_type_existing_ctrl):
            logger.info("Controller " + index + " has same type and matching parameters like " +
                        "controllers " + str(['%i' % idx for idx in same_type_existing_ctrl]))
    else:
        logger.info("Creating controller " + index + " of type %s " % this_ctrl_type)
        logger.debug("no matching parameters are given to check whether problematic, " +
                     "same type controllers already exist.")


def drop_same_type_existing_controllers(net, this_ctrl_type, index=None, matching_params=None,
                                        **kwargs):
    """
    Drops same type controllers to create a new controller of this type.
    INPUT:
        **net** - pandapower net

        **this_ctrl_type** (controller object or string name of controller object)

    OPTIONAL:
        **index** (int) - index in net.controller of the controller to be created

        **matching_params** (dict) - parameters, which must be equal if same type controller should
            be logged.

        ****kwargs** - unused arguments, given to avoid unexpected input arguments
    """
    index = str(index)
    if isinstance(matching_params, dict):
        same_type_existing_ctrl = get_controller_index(net, ctrl_type=this_ctrl_type,
                                                       parameters=matching_params)
        if len(same_type_existing_ctrl):
            net.controller = net.controller.drop(same_type_existing_ctrl)
            logger.debug("Controllers " + str(['%i' % idx for idx in same_type_existing_ctrl]) +
                         "got removed because of same type and matching parameters as new " +
                         "controller " + index + ".")
    else:
        logger.info("Creating controller " + index + " of type %s, " % this_ctrl_type +
                    "no matching parameters are given to check which " +
                    "same type controllers should be dropped.")


def plot_characteristic(characteristic, start, stop, num=20, xlabel=None, ylabel=None):
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    x = np.linspace(start, stop, num)
    y = characteristic(x)
    plt.plot(x, y, marker='x')
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)


def create_trafo_characteristic_object(net):
    # check if trafo_characteristic_spline table already exists & if so, delete & re-create
    if "trafo_characteristic_spline" in net:
        del net["trafo_characteristic_spline"]
    # 2-winding transformers
    if (net['trafo_characteristic_table'].index.size > 0 and
            net['trafo']['id_characteristic_table'].notna().any()):
        time_start = time.time()
        logger.info("Creating tap dependent characteristic objects for 2w-trafos.")
        characteristic_df_temp = net['trafo_characteristic_table'][
            ['id_characteristic', 'step', 'voltage_ratio', 'angle_deg', 'vk_percent', 'vkr_percent']]
        for trafo_id, trafo_row in net.trafo.dropna(subset=['id_characteristic_table']).iterrows():
            characteristic_df = characteristic_df_temp.loc[
                characteristic_df_temp['id_characteristic'] == trafo_row['id_characteristic_table']]
            variables = ['voltage_ratio', 'angle_deg', 'vk_percent', 'vkr_percent']
            variables_filtered = [var for var in variables if var in characteristic_df.columns]
            x_points = [characteristic_df['step'].to_list()]
            y_points = {col: [characteristic_df[col].tolist()] for col in variables_filtered}
            _create_trafo_characteristics(net, "trafo", [trafo_id], variables_filtered,
                                          x_points, y_points)
        logger.info(f"Finished creating tap dependent characteristic objects for 2w-trafos in "
                    f"{time.time() - time_start}.")
    else:
        logger.info("trafo_characteristic_table has no values for 2w-trafos - no characteristic objects created.")
    # 3-winding transformers
    if (net['trafo_characteristic_table'].index.size > 0 and
            net['trafo3w']['id_characteristic_table'].notna().any()):
        time_start = time.time()
        logger.info("Creating tap dependent characteristic objects for 3w-trafos.")
        characteristic_df_temp = net['trafo_characteristic_table'][
                ['id_characteristic', 'step', 'voltage_ratio', 'angle_deg', 'vk_hv_percent', 'vkr_hv_percent',
                 'vk_mv_percent', 'vkr_mv_percent', 'vk_lv_percent', 'vkr_lv_percent']]
        for trafo_id, trafo_row in net.trafo3w.dropna(subset=['id_characteristic_table']).iterrows():
            characteristic_df = characteristic_df_temp.loc[
                characteristic_df_temp['id_characteristic'] == trafo_row['id_characteristic_table']]
            variables = ['voltage_ratio', 'angle_deg', 'vk_hv_percent', 'vkr_hv_percent', 'vk_mv_percent',
                         'vkr_mv_percent', 'vk_lv_percent', 'vkr_lv_percent']
            variables_filtered = [var for var in variables if var in characteristic_df.columns]
            x_points = [characteristic_df['step'].to_list()]
            y_points = {col: [characteristic_df[col].tolist()] for col in variables_filtered}
            _create_trafo_characteristics(net, "trafo3w", [trafo_id], variables_filtered,
                                          x_points, y_points)
        logger.info(f"Finished creating tap dependent characteristic objects for 3w-trafos in "
                    f"{time.time() - time_start}.")
    else:
        logger.info("trafo_characteristic_table has no values for 3w-trafos - no characteristic objects created.")

    # pivot spline characteristic objects to have one row per trafo/trafo3w
    net["trafo_characteristic_spline"] = net["trafo_characteristic_spline_temp"].applymap(
        lambda x: net["trafo_characteristic_spline"].loc[x, 'object'] if pd.notna(x) else pd.NA).sort_index()
    # create id_characteristic column
    net["trafo_characteristic_spline"]["id_characteristic"] = net["trafo_characteristic_spline"].index
    net["trafo_characteristic_spline"].insert(
        0, 'id_characteristic', net["trafo_characteristic_spline"].pop('id_characteristic'))
    del net["trafo_characteristic_spline_temp"]


def _create_trafo_characteristics(net, trafotable, trafo_index, variable, x_points, y_points):
    supported_columns = {"trafo": ["voltage_ratio_characteristic", "angle_deg_characteristic",
                                   "vk_percent_characteristic", "vkr_percent_characteristic"],
                         "trafo3w": ["voltage_ratio_characteristic", "angle_deg_characteristic",
                                     "vk_hv_percent_characteristic", "vkr_hv_percent_characteristic",
                                     "vk_mv_percent_characteristic", "vkr_mv_percent_characteristic",
                                     "vk_lv_percent_characteristic", "vkr_lv_percent_characteristic"]}

    # create or re-populate id_characteristic_spline column - same indices as id_characteristic_table
    net[trafotable]["id_characteristic_spline"] = net[trafotable]["id_characteristic_table"].copy()
    if "trafo_characteristic_spline_temp" not in net:
        col_list = list(dict.fromkeys(supported_columns["trafo"] + supported_columns["trafo3w"]))
        net["trafo_characteristic_spline_temp"] = pd.DataFrame(columns=col_list, dtype="Int64")

    for var in variable:
        # create characteristics for the specified variable and set their indices in the trafo table
        col = f"{var}_characteristic"
        # check if the variable is a valid attribute of the trafo table
        if col not in supported_columns[trafotable]:
            raise UserWarning("Variable %s is not supported for table %s" % (var, trafotable))

        # check inputs, check if 1 trafo or multiple, verify shape of x_points and y_points and trafo_index
        if hasattr(trafo_index, '__iter__'):
            single_mode = False
            if not (len(trafo_index) == len(x_points) == len(y_points[var])):
                raise UserWarning("The lengths of the trafo index and points do not match!")
        else:
            single_mode = True
            if len(x_points) != len(y_points[var]):
                raise UserWarning("The lengths of the points do not match!")

        if single_mode:
            zip_params = zip([trafo_index], [x_points], [y_points[var]])
        else:
            zip_params = zip(trafo_index, x_points, y_points[var])

        for tid, x_p, y_p in zip_params:
            # create the characteristic and set its index in the trafotable
            s = SplineCharacteristic(net, x_p, y_p, table="trafo_characteristic_spline")
            idx = net[trafotable].at[tid, "id_characteristic_spline"]
            # save the index of the new spline characteristic object in the temp table
            net["trafo_characteristic_spline_temp"].at[idx, col] = s.index


def create_shunt_characteristic_object(net):
    # check if shunt_characteristic_spline table already exists & if so, delete & re-create
    if "shunt_characteristic_spline" in net:
        del net["shunt_characteristic_spline"]
    if net['shunt_characteristic_table'].index.size > 0:
        time_start = time.time()
        logger.info("Creating step dependent power characteristic objects for shunts.")
        characteristic_df_temp = net['shunt_characteristic_table']
        for shunt_id, shunt_row in net.shunt.dropna(subset=['id_characteristic_table']).iterrows():
            characteristic_df = characteristic_df_temp.loc[
                characteristic_df_temp['id_characteristic'] == shunt_row['id_characteristic_table']]
            variables = ['q_mvar', 'p_mw']
            variables_filtered = [var for var in variables if var in characteristic_df.columns]
            x_points = [characteristic_df['step'].to_list()]
            y_points = {col: [characteristic_df[col].tolist()] for col in variables_filtered}
            _create_shunt_characteristics(net, [shunt_id], variables_filtered,
                                          x_points, y_points)
        logger.info(f"Finished creating step dependent power characteristic objects for shunts in"
                    f"{time.time() - time_start}.")
    else:
        logger.info("shunt_characteristic_table is empty - no characteristic objects created.")

    # pivot spline characteristic objects to have one row per shunt
    net["shunt_characteristic_spline"] = net["shunt_characteristic_spline_temp"].applymap(
        lambda x: net["shunt_characteristic_spline"].loc[x, 'object'] if pd.notna(x) else pd.NA).sort_index()
    # create id_characteristic column
    net["shunt_characteristic_spline"]["id_characteristic"] = net["shunt_characteristic_spline"].index
    net["shunt_characteristic_spline"].insert(
        0, 'id_characteristic', net["shunt_characteristic_spline"].pop('id_characteristic'))
    del net["shunt_characteristic_spline_temp"]


def _create_shunt_characteristics(net, shunt_index, variable, x_points, y_points):
    supported_columns = ["q_mvar_characteristic", "p_mw_characteristic"]

    # create id_characteristic_spline column - same indices as id_characteristic_table
    if "id_characteristic_spline" not in net["shunt"]:
        net["shunt"]["id_characteristic_spline"] = net["shunt"]["id_characteristic_table"].copy()
    if "shunt_characteristic_spline_temp" not in net:
        net["shunt_characteristic_spline_temp"] = pd.DataFrame(columns=supported_columns, dtype="Int64")

    for var in variable:
        # create characteristics for the specified variable and set their indices in the shunt table
        col = f"{var}_characteristic"
        # check if the variable is a valid attribute of the shunt table
        if col not in supported_columns:
            raise UserWarning("Variable %s is not supported for table shunt" % var)

        # check inputs, check if 1 shunt or multiple, verify shape of x_points and y_points and shunt_index
        if hasattr(shunt_index, '__iter__'):
            single_mode = False
            if not (len(shunt_index) == len(x_points) == len(y_points[var])):
                raise UserWarning("The lengths of the shunt index and points do not match!")
        else:
            single_mode = True
            if len(x_points) != len(y_points[var]):
                raise UserWarning("The lengths of the points do not match!")

        if single_mode:
            zip_params = zip([shunt_index], [x_points], [y_points[var]])
        else:
            zip_params = zip(shunt_index, x_points, y_points[var])

        for tid, x_p, y_p in zip_params:
            # create the characteristic and set its index in the shunt table
            s = SplineCharacteristic(net, x_p, y_p, table="shunt_characteristic_spline")
            idx = net["shunt"].at[tid, "id_characteristic_spline"]
            # save the index of the new spline characteristic object in the temp table
            net["shunt_characteristic_spline_temp"].at[idx, col] = s.index


def _set_reactive_capability_curve_flag(net, element):
    if element not in ["gen", "sgen"]:
        UserWarning(f"The given {element} type is not valid for setting curve dependency table flag. "
                    f"Please give gen or sgen as an argument of the function")
        return
    # Quick checks for element table and required columns
    if (len(net[element]) == 0 or
            not {"id_q_capability_curve_characteristic", "reactive_capability_curve", "curve_style"}.issubset(net[element].columns)
            or (not net[element]['id_q_capability_curve_characteristic'].notna().any() and
                not net[element]['reactive_capability_curve'].any()) and not net[element]['curve_style'].any()):
        logger.info(f"No {element} with Q capability curve table found.")
    else:
        net[element]['reactive_capability_curve'] = (
                net[element]['id_q_capability_curve_characteristic'].notna() &
                (net[element]['id_q_capability_curve_characteristic'] >= 0) &
                net[element]['curve_style'].isin(["straightLineYValues", "constantYValue"])
        ).astype(bool)


def create_q_capability_curve_characteristics_object(net):
    # check if element_characteristic_spline table already exists & if so, delete & re-create
    if "q_capability_curve_characteristic" in net:
        del net["q_capability_curve_characteristic"]

    # create characteristics
    if "q_capability_curve_table" in net.keys() and net['q_capability_curve_table'].index.size > 0:
        time_start = time.time()

        # Set flag reactive_capability_curve
        _set_reactive_capability_curve_flag(net, "gen")
        _set_reactive_capability_curve_flag(net, "sgen")

        # Create Q capability curve characteristics dataframe
        net["q_capability_curve_characteristic"] = pd.DataFrame({
            "id_q_capability_curve": pd.Series(dtype="Int64"),
            "q_min_characteristic": pd.Series(dtype="object"),
            "q_max_characteristic": pd.Series(dtype="object"),
        })

        net["q_capability_curve_characteristic_temp"] = pd.DataFrame()
        characteristic_df_temp = net['q_capability_curve_table']
        mydata_grouped = characteristic_df_temp.groupby('id_q_capability_curve')
        net["q_capability_curve_characteristic"]["id_q_capability_curve"] = mydata_grouped.size().index

        # Prepare lists
        element_ids = []
        q_min_indices = []
        q_max_indices = []

        # Iterate directly over grouped data
        for element_id, group_data in mydata_grouped:
            p_mw_values = np.hstack(group_data['p_mw'])
            q_min_values = np.hstack(group_data['q_min_mvar'])
            q_max_values = np.hstack(group_data['q_max_mvar'])

            # Compute Characteristic indices
            q_min_index = Characteristic(
                net, p_mw_values, q_min_values, table="q_capability_curve_characteristic_temp").index
            q_max_index = Characteristic(
                net, p_mw_values, q_max_values, table="q_capability_curve_characteristic_temp").index

            # Collect results
            element_ids.append(element_id)
            q_min_indices.append(q_min_index)
            q_max_indices.append(q_max_index)
            logger.info("Adding characteristic objects for id_q_capability_curve %d" % element_id)

        characteristic_df = pd.DataFrame({
            "id_q_capability_curve": element_ids,
            "q_min_characteristic": q_min_indices,
            "q_max_characteristic": q_max_indices,
        }).set_index("id_q_capability_curve")

        net["q_capability_curve_characteristic"] = net[
            "q_capability_curve_characteristic"].combine_first(characteristic_df)

        # Extract the temporary table containing the objects
        temp_table = net["q_capability_curve_characteristic_temp"]

        # Map the indices in `q_min_index` and `q_max_index` to the corresponding objects in the temporary table
        if not temp_table.empty:
            object_map = temp_table["object"]  # Cache the mapping for efficiency
            net["q_capability_curve_characteristic"]["q_min_characteristic"] = net[
                "q_capability_curve_characteristic"]["q_min_characteristic"].map(object_map)
            net["q_capability_curve_characteristic"]["q_max_characteristic"] = net[
                "q_capability_curve_characteristic"]["q_max_characteristic"].map(object_map)
        logger.info(f"Finished creating p dependent q characteristic objects for capability curve in "
                    f"{time.time() - time_start}.")
        del net["q_capability_curve_characteristic_temp"]

    else:
        logger.info("q_capability_curve_table is empty - no characteristic objects created.")



# -*- coding: utf-8 -*-
from pandapower.pypower.idx_brch import branch_cols
# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.pypower.idx_bus import VM, VA
from pandapower.pypower.idx_gen import GEN_BUS, GEN_STATUS, VG
from pandapower.pypower.bustypes import bustypes
from numpy import flatnonzero as find, pi, exp, int64, hstack, zeros, float64

def _get_pf_variables_from_ppci(ppci, vsc_ref=False):
    ## default arguments
    if ppci is None:
        ValueError('ppci is empty')
    # ppopt = ppoption(ppopt)

    # get data for calc
    bus, gen, vsc = ppci["bus"], ppci["gen"], ppci["vsc"]

    # if ppc["branch"] comes from the pypower -> pandapower converter, it has fewer columns than ppc in pandapower
    # because it is lacking BR_R_ASYM, BR_X_ASYM, BR_G, BR_G_ASYM, BR_B_ASYM, and it is OK to use 0 as default values
    branch = ppci["branch"]
    br_shape = branch.shape
    if br_shape[1] < branch_cols:
        branch = hstack([branch, zeros(shape=(br_shape[0], branch_cols - br_shape[1]), dtype=float64)])

    ## get bus index lists of each type of bus
    ref, pv, pq = bustypes(bus, gen, vsc if vsc_ref else None)

    ## generator info
    on = find(gen[:, GEN_STATUS] > 0)  ## which generators are on?
    gbus = gen[on, GEN_BUS].astype(int64)  ## what buses are they at?

    ## initial state
    # V0    = ones(bus.shape[0])            ## flat start
    V0 = bus[:, VM] * exp(1j * pi / 180. * bus[:, VA])
    V0[gbus] = gen[on, VG] / abs(V0[gbus]) * V0[gbus]

    ref_gens = ppci["internal"]["ref_gens"]
    return ppci["baseMVA"], bus, gen, branch, ppci["svc"], ppci["tcsc"], ppci["ssc"], vsc, \
        ref, pv, pq, on, gbus, V0, ref_gens


def _store_results_from_pf_in_ppci(ppci, bus, gen, branch, success, iterations, et):
    ppci["bus"], ppci["gen"], ppci["branch"] = bus, gen, branch
    ppci["success"] = bool(success)
    ppci["iterations"] = iterations
    ppci["et"] = et
    return ppci



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.auxiliary import ensure_iterability
from pandapower.control.controller.DERController.DERBasics import BaseModel, QVCurve, \
    CosphiVCurve, CosphiPCurve

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


class QModel(BaseModel):
    """
    Base class to model how q is determined in step().
    """

    def __init__(self, **kwargs):
        pass

    def step(self, vm_pu=None, p_pu=None):
        pass

    def __str__(self):
        return self.__class__.name()


class QModelConstQ(QModel):
    """
    Class to model a fixed q value.
    """

    def __init__(self, q_pu):
        self.q_pu = q_pu

    def step(self, vm_pu=None, p_pu=None):
        if p_pu is not None:
            len_ = len(p_pu)
        elif vm_pu is not None:
            len_ = len(vm_pu)
        else:
            len_ = None
        return np.array(ensure_iterability(self.q_pu, len_))

    def __str__(self):
        return self.__class__.name() + f"(q={self.q_pu})"


class QModelCosphiVCurve(QModel):
    """
    Base class to model that q is determined in dependency of the voltage and active power.
    Class to model that q is determined in dependency of a voltage dependent cosphi curve.
    """

    def __init__(self, cosphi_v_curve):
        if isinstance(cosphi_v_curve, dict):
            self.cosphi_v_curve = CosphiVCurve(**cosphi_v_curve)
        else:
            self.cosphi_v_curve = cosphi_v_curve

    def step(self, vm_pu, p_pu=None):
        assert p_pu is not None
        assert all(p_pu >= 0)
        return self.cosphi_v_curve.step(vm_pu, p_pu)


class QModelCosphiP(QModel):
    """
    Class to model that q is determined in dependency of the momentary active power by a fixed
    cosphi. q is calculated based on: `q = sin(phi) * p` instead of `q = sin(phi) * s` -> It ignores
    that s changes by changing q.
    """

    def __init__(self, cosphi: float):
        assert -1 <= cosphi <= 1
        self.cosphi = cosphi
        self.q_setpoint_pu = np.sign(self.cosphi) * np.sqrt(1-self.cosphi**2)

    def step(self, vm_pu=None, p_pu=None,):
        assert p_pu is not None
        if any(p_pu < 0):
            logger.warning("p < 0 is assumed as p=0 in QModelCosphiP.step()")
            p_pu[p_pu< 0] = 0
        return (p_pu * self.q_setpoint_pu)

    def __str__(self):
        return self.__class__.name() + f"(cosphi={self.cosphi})"


class QModelCosphiSn(QModel):
    """
    Class to model that q is determined in dependency of the apparent power by a fixed cosphi. q is
    calculated based on: `q = sin(phi) * s` only considering sn_mva and ignoring that s changes with
    p and q.
    """

    def __init__(self, cosphi=0.2):
        self.cosphi = cosphi

    def step(self, vm_pu=None, p_pu=None):
        return (1 * self.cosphi)

    def __str__(self):
        return self.__class__.name() + f"(cosphi={self.cosphi})"


class QModelCosphiPQ(QModel):
    """
    Class to model that q is determined in dependency of the momentary active power by a fixed
    cosphi. q is calculated based on: `q = sin(phi) * s` with `s = (p**2 + q**2)**0.5` -> the
    actual cosphi covers the requested cosphi.
    """

    def __init__(self, cosphi: float):
        assert -1 <= cosphi <= 1
        self.cosphi = cosphi
        self.q_setpoint_pu = np.sign(self.cosphi) * np.sqrt(1-self.cosphi**2)

    def step(self, vm_pu=None, p_pu=None,):
        assert p_pu is not None
        if any(p_pu < 0):
            logger.warning("p_pu < 0 is assumed as p_pu=0 QModelCosphiPQ.step()")
            p_pu[p_pu < 0] = 0
        return (p_pu/abs(self.cosphi) * self.q_setpoint_pu)

    def __str__(self):
        return self.__class__.name() + f"(cosphi={self.cosphi})"


class QModelCosphiPCurve(QModel):
    """
    Class to model that q is determined in dependency of a cosphi which depends on the active power.
    """

    def __init__(self, cosphi_p_curve):
        if isinstance(cosphi_p_curve, dict):
            self.cosphi_p_curve = CosphiPCurve(**cosphi_p_curve)
        else:
            self.cosphi_p_curve = cosphi_p_curve

    def step(self, vm_pu=None, p_pu=None):
        assert p_pu is not None
        assert all(p_pu >= 0)
        return self.cosphi_p_curve.step(p_pu)


class QModelQVCurve(QModel):
    """
    Base class to model that q is determined in dependency of the voltage.
    """

    def __init__(self, qv_curve):
        if isinstance(qv_curve, dict):
            self.qv_curve = QVCurve(**qv_curve)
        else:
            self.qv_curve = qv_curve

    def step(self, vm_pu, p_pu=None):
        q_pu = self.qv_curve.step(vm_pu)
        return q_pu


if __name__ == "__main__":
    pass
